# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["OptionModifier", "OptionLike", "CommandLike", "Command", "ApplicationCommand", "SubCommandTrait"]

import abc
import asyncio
import collections
import dataclasses
import datetime
import enum
import inspect
import re
import typing as t

import hikari

from lightbulb import errors

if t.TYPE_CHECKING:
    from lightbulb import app as app_
    from lightbulb import buckets
    from lightbulb import checks
    from lightbulb import context as context_
    from lightbulb import cooldowns
    from lightbulb import events
    from lightbulb import parser as parser_
    from lightbulb import plugins

_AutocompleteableOptionT = t.Union[str, int, float]
AutocompleteCallbackT = t.TypeVar(
    "AutocompleteCallbackT",
    bound=t.Callable[
        ...,
        t.Coroutine[
            t.Any,
            t.Any,
            t.Union[
                _AutocompleteableOptionT,
                hikari.api.AutocompleteChoiceBuilder,
                t.Sequence[t.Union[_AutocompleteableOptionT, hikari.api.AutocompleteChoiceBuilder]],
            ],
        ],
    ],
)
ErrorHandlerCallbackT = t.TypeVar(
    "ErrorHandlerCallbackT", bound=t.Callable[..., t.Coroutine[t.Any, t.Any, t.Optional[bool]]]
)


class CommandCallbackT(t.Protocol):
    def __call__(self, context: context_.base.Context, **kwargs: t.Any) -> t.Coroutine[t.Any, t.Any, None]:
        ...


OPTION_TYPE_MAPPING = {
    str: hikari.OptionType.STRING,
    int: hikari.OptionType.INTEGER,
    float: hikari.OptionType.FLOAT,
    bool: hikari.OptionType.BOOLEAN,
    hikari.User: hikari.OptionType.USER,
    hikari.Member: hikari.OptionType.USER,
    hikari.GuildChannel: hikari.OptionType.CHANNEL,
    hikari.TextableGuildChannel: hikari.OptionType.CHANNEL,
    hikari.TextableChannel: hikari.OptionType.CHANNEL,
    hikari.GuildCategory: hikari.OptionType.CHANNEL,
    hikari.GuildVoiceChannel: hikari.OptionType.CHANNEL,
    hikari.Role: hikari.OptionType.ROLE,
    hikari.Emoji: hikari.OptionType.STRING,
    hikari.Guild: hikari.OptionType.STRING,
    hikari.Message: hikari.OptionType.STRING,
    hikari.Invite: hikari.OptionType.STRING,
    hikari.Colour: hikari.OptionType.STRING,
    hikari.Color: hikari.OptionType.STRING,
    hikari.Snowflake: hikari.OptionType.STRING,
    datetime.datetime: hikari.OptionType.STRING,
    hikari.Attachment: hikari.OptionType.ATTACHMENT,
}
OPTION_NAME_REGEX: re.Pattern[str] = re.compile(r"^[\w-]{1,32}$", re.U)


class _HasRecreateSubcommands(t.Protocol):
    app: app_.BotApp

    def recreate_subcommands(self, raw_cmds: t.Sequence[CommandLike], app: app_.BotApp) -> None:
        ...


class _SubcommandListProxy(collections.UserList):  # type: ignore
    __slots__ = ("parents",)

    def __init__(self, *args: t.Any, parent: _HasRecreateSubcommands, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self.parents = [parent]

    def append(self, item: t.Any) -> None:
        super().append(item)
        for parent in self.parents:
            parent.recreate_subcommands(self.data, parent.app)

    def add_parent(self, parent: _HasRecreateSubcommands) -> _SubcommandListProxy:
        self.parents.append(parent)
        return self


def _get_choice_objects_from_choices(
    choices: t.Sequence[t.Union[str, int, float, hikari.CommandChoice]]
) -> t.Sequence[hikari.CommandChoice]:
    return [c if isinstance(c, hikari.CommandChoice) else hikari.CommandChoice(name=str(c), value=c) for c in choices]


class OptionModifier(enum.Enum):
    """Enum representing option modifiers that affect parsing for prefix commands."""

    NONE = enum.auto()
    """No modifier. This will be parsed as a normal argument."""
    GREEDY = enum.auto()
    """Greedy option. This will consume arguments until the string is exhausted or conversion fails."""
    CONSUME_REST = enum.auto()
    """Consume rest option. This will consume the entire remainder of the string."""


@dataclasses.dataclass
class OptionLike:
    """
    Generic dataclass representing a command option. Compatible with both prefix and application commands.
    """

    name: str
    """The name of the option."""
    description: str
    """The description of the option"""
    arg_type: t.Any = str
    """The type of the option."""
    required: bool = True
    """Whether or not the option is required. This will be inferred from whether or not a default value was provided if unspecified."""
    choices: t.Optional[t.Sequence[t.Union[str, int, float, hikari.CommandChoice]]] = None
    """The option's choices. This only affects slash commands."""
    channel_types: t.Optional[t.Sequence[hikari.ChannelType]] = None
    """The channel types for this option. This only affects slash commands."""
    default: hikari.UndefinedOr[t.Any] = hikari.UNDEFINED
    """The default value for this option."""
    modifier: OptionModifier = OptionModifier.NONE
    """Additional modifier controlling how the option should be parsed. This only affects prefix commands."""
    min_value: t.Optional[t.Union[float, int]] = None
    """
    The minimum value permitted for this option (inclusive). The option must be ``INTEGER`` or ``FLOAT`` to use this.

    .. versionadded:: 2.1.3
    """
    max_value: t.Optional[t.Union[float, int]] = None
    """
    The maximum value permitted for this option (inclusive). The option must be ``INTEGER`` or ``FLOAT`` to use this.

    .. versionadded:: 2.1.3
    """
    min_length: t.Optional[int] = None
    """
    The minimum length permitted for this option. The option must be ``STRING`` to use this.

    .. versionadded:: 2.3.2
    """
    max_length: t.Optional[int] = None
    """
    The maximum length permitted for this option. The option must be ``STRING`` to use this.

    .. versionadded:: 2.3.2
    """
    autocomplete: bool = False
    """Whether the option should be autocompleted or not. This only affects slash commands."""
    name_localizations: t.Mapping[t.Union[hikari.Locale, str], str] = dataclasses.field(default_factory=dict)
    """
    A mapping of locale to name localizations for this option

    .. versionadded:: 2.3.0
    """
    description_localizations: t.Mapping[t.Union[hikari.Locale, str], str] = dataclasses.field(default_factory=dict)
    """
    A mapping of locale to description localizations for this option

    .. versionadded:: 2.3.0
    """

    def as_application_command_option(self) -> hikari.CommandOption:
        """
        Convert this object into a :obj:`~hikari.commands.CommandOption`.

        Returns:
            :obj:`~hikari.commands.CommandOption`: Created ``CommandOption`` object.
        """
        if not OPTION_NAME_REGEX.fullmatch(self.name) or self.name != self.name.lower():
            raise ValueError(
                f"Application command option {self.name!r}: name must match regex '^[\\w-]{1,32}$' and be all lowercase"
            )
        if len(self.description) < 1 or len(self.description) > 100:
            raise ValueError(
                f"Application command option {self.name!r}: description must be from 1-100 characters long"
            )

        arg_type = OPTION_TYPE_MAPPING.get(self.arg_type, self.arg_type)
        if not isinstance(arg_type, hikari.OptionType):
            arg_type = hikari.OptionType.STRING  # type: ignore[unreachable]

        if (self.min_value is not None or self.max_value is not None) and arg_type not in (
            hikari.OptionType.INTEGER,
            hikari.OptionType.FLOAT,
        ):
            raise ValueError(
                f"Application command option {self.name!r}: 'min_value' or 'max_value' was provided but the option type is not numeric"
            )

        if (self.min_length is not None or self.max_length is not None) and arg_type is not hikari.OptionType.STRING:
            raise ValueError(
                f"Application command option {self.name!r}: 'min_length' or 'max_length' was provided but the option type is not string"
            )

        if (
            arg_type not in (hikari.OptionType.INTEGER, hikari.OptionType.FLOAT, hikari.OptionType.STRING)
            and self.autocomplete
        ):
            raise ValueError(
                f"Application command option {self.name!r}: 'autocomplete' is True but the option type does not support choices"
            )

        kwargs: t.MutableMapping[str, t.Any] = {
            "type": arg_type,
            "name": self.name,
            "description": self.description,
            "is_required": self.required,
            "autocomplete": self.autocomplete,
            "name_localizations": self.name_localizations,
            "description_localizations": self.description_localizations,
        }

        if self.choices:
            if len(self.choices) > 25:
                raise ValueError("Application command options can have at most 25 choices")
            kwargs["choices"] = _get_choice_objects_from_choices(self.choices)

        if self.channel_types:
            kwargs["channel_types"] = self.channel_types

        if self.min_value is not None:
            kwargs["min_value"] = (
                int(self.min_value) if arg_type is hikari.OptionType.INTEGER else float(self.min_value)
            )

        if self.max_value is not None:
            kwargs["max_value"] = (
                int(self.max_value) if arg_type is hikari.OptionType.INTEGER else float(self.max_value)
            )

        if self.min_length is not None:
            kwargs["min_length"] = self.min_length
        if self.max_length is not None:
            kwargs["max_length"] = self.max_length

        return hikari.CommandOption(**kwargs)


@dataclasses.dataclass
class CommandLike:
    """Generic dataclass representing a command. This can be converted into any command object."""

    callback: CommandCallbackT
    """The callback function for the command."""
    name: str
    """The name of the command."""
    description: str
    """The description of the command."""
    options: t.MutableMapping[str, OptionLike] = dataclasses.field(default_factory=dict)
    """The options for the command."""
    checks: t.Sequence[t.Union[checks.Check, checks._ExclusiveCheck]] = dataclasses.field(default_factory=list)
    """The checks for the command."""
    error_handler: t.Optional[
        t.Callable[[events.CommandErrorEvent], t.Coroutine[t.Any, t.Any, t.Optional[bool]]]
    ] = None
    """The error handler for the command."""
    aliases: t.Sequence[str] = dataclasses.field(default_factory=list)
    """The aliases for the command. This only affects prefix commands."""
    guilds: hikari.UndefinedOr[t.Sequence[int]] = hikari.UNDEFINED
    """The guilds for the command. This only affects application commands."""
    subcommands: t.List[CommandLike] = dataclasses.field(default_factory=list)
    """Subcommands for the command."""
    parser: t.Optional[t.Type[parser_.BaseParser]] = None
    """The argument parser to use for prefix commands."""
    cooldown_manager: t.Optional[cooldowns.CooldownManager] = None
    """The cooldown manager for the command."""
    help_getter: t.Optional[t.Callable[[Command, context_.base.Context], str]] = None
    """The function to call to get the command's long help text."""
    auto_defer: bool = False
    """Whether or not to automatically defer the response when the command is invoked."""
    ephemeral: bool = False
    """Whether or not to send responses from this command as ephemeral messages by default."""
    check_exempt: t.Optional[t.Callable[[context_.base.Context], t.Union[bool, t.Coroutine[t.Any, t.Any, bool]]]] = None
    """Check exempt predicate to use for the command."""
    hidden: bool = False
    """Whether or not the command should be hidden from the help command."""
    inherit_checks: bool = False
    """Whether or not the command should inherit checks from the parent group."""
    pass_options: bool = False
    """
    Whether or not the command will have its options passed as keyword arguments when invoked.

    .. versionadded:: 2.2.1
    """
    max_concurrency: t.Optional[t.Tuple[int, t.Type[buckets.Bucket]]] = None
    """
    The max concurrency rule for the command.

    .. versionadded:: 2.2.1
    """
    app_command_default_member_permissions: t.Optional[hikari.Permissions] = None
    """
    The default member permissions for this command, if an application command.

    .. versionadded:: 2.2.3
    """
    app_command_dm_enabled: bool = True
    """
    Whether this command will be enabled in DMs, if an application command.

    .. versionadded:: 2.2.3
    """
    app_command_bypass_author_permission_checks: bool = False
    """
    Whether invocations of this command will bypass author permission checks, if an application command.

    .. versionadded:: 2.2.3
    """
    name_localizations: t.Mapping[t.Union[hikari.Locale, str], str] = dataclasses.field(default_factory=dict)
    """
    A mapping of locale to name localizations for this command

    .. versionadded:: 2.3.0
    """
    description_localizations: t.Mapping[t.Union[hikari.Locale, str], str] = dataclasses.field(default_factory=dict)
    """
    A mapping of locale to description localizations for this command

    .. versionadded:: 2.3.0
    """
    nsfw: bool = False
    """
    Whether the command should only be enabled in NSFW channels.

    For prefix commands, this will add an NSFW-channel only check to the command automatically.
    For slash commands, this will behave as specified in the Discord documentation.

    .. versionadded:: 2.3.1
    """
    _autocomplete_callbacks: t.Dict[
        str,
        t.Callable[
            [hikari.CommandInteractionOption, hikari.AutocompleteInteraction],
            t.Coroutine[
                t.Any,
                t.Any,
                t.Union[
                    _AutocompleteableOptionT,
                    hikari.api.AutocompleteChoiceBuilder,
                    t.Sequence[t.Union[_AutocompleteableOptionT, hikari.api.AutocompleteChoiceBuilder]],
                ],
            ],
        ],
    ] = dataclasses.field(default_factory=dict, init=False)

    async def __call__(self, context: context_.base.Context) -> None:
        await self.callback(context)

    @t.overload
    def set_error_handler(self) -> t.Callable[[ErrorHandlerCallbackT], ErrorHandlerCallbackT]:
        ...

    @t.overload
    def set_error_handler(self, func: ErrorHandlerCallbackT) -> ErrorHandlerCallbackT:
        ...

    def set_error_handler(
        self,
        func: t.Optional[ErrorHandlerCallbackT] = None,
    ) -> t.Union[ErrorHandlerCallbackT, t.Callable[[ErrorHandlerCallbackT], ErrorHandlerCallbackT]]:
        """
        Registers a coroutine function as an error handler for this command. This can be used as a first or second
        order decorator, or called manually with the function to register.

        Example:

            .. code-block:: python

                @lightbulb.command(...)
                @lightbulb.implements(...)
                async def foo(ctx: lightbulb.Context) -> None:
                    ...

                # Valid
                @foo.set_error_handler
                async def foo_error_handler(event: lightbulb.CommandErrorEvent) -> bool:
                    ...

                # Also valid
                @foo.set_error_handler()
                async def foo_error_handler(event: lightbulb.CommandErrorEvent) -> bool:
                    ...

                # Also valid
                async def foo_error_handler(event: lightbulb.CommandErrorEvent) -> bool:
                    ...

                foo.set_error_handler(foo_error_handler)
        """
        if func is not None:
            self.error_handler = func
            return func

        def decorate(func_: ErrorHandlerCallbackT) -> ErrorHandlerCallbackT:
            self.error_handler = func_
            return func_

        return decorate

    @t.overload
    def child(self) -> t.Callable[[CommandLike], CommandLike]:
        ...

    @t.overload
    def child(self, cmd_like: CommandLike) -> CommandLike:
        ...

    def child(
        self, cmd_like: t.Optional[CommandLike] = None
    ) -> t.Union[CommandLike, t.Callable[[CommandLike], CommandLike]]:
        """
        Registers a :obj:`~CommandLike` object as a child to this command. This can be used as a first or second
        order decorator, or called manually with the :obj:`~CommandLike` instance to add as a child.

        Example:

            .. code-block:: python

                @lightbulb.command(...)
                @lightbulb.implements(...)
                async def foo(ctx: lightbulb.Context) -> None:
                    ...

                # Valid
                @foo.child
                @lightbulb.command(...)
                @lightbulb.implements(...)
                async def foo_child(event: lightbulb.Context) -> None:
                    ...

                # Also valid
                @foo.child()
                @lightbulb.command(...)
                @lightbulb.implements(...)
                async def foo_child(event: lightbulb.Context) -> None:
                    ...

                # Also valid
                @lightbulb.command(...)
                @lightbulb.implements(...)
                async def foo_child(event: lightbulb.Context) -> None:
                    ...

                foo.child(foo_child)
        """
        if cmd_like is not None:
            self.subcommands.append(cmd_like)
            return cmd_like

        def decorate(cmd_like_: CommandLike) -> CommandLike:
            self.subcommands.append(cmd_like_)
            return cmd_like_

        return decorate

    def autocomplete(self, opt1: str, *opts: str) -> t.Callable[[AutocompleteCallbackT], AutocompleteCallbackT]:
        """
        Second order decorator that registers a function as an autocomplete callback for this command.

        The autocomplete callback **must** be an asynchronous function that takes exactly two arguments: ``option``
        (an instance of :obj:`hikari.interactions.command_interactions.AutocompleteInteractionOption`) which is
        the option being autocompleted, and ``interaction`` (an instance of :obj:`hikari.interactions.command_interactions.AutocompleteInteraction`)
        which is the interaction that triggered the autocomplete.

        Autocomplete can only be enabled for options with type :obj:`str`, :obj:`int`, or :obj:`float`.

        The callback should return one of the following: a single item of the option type, a sequence of items of the
        option type, a single :obj:`hikari.commands.CommandChoice`, or a sequence of :obj:`hikari.commands.CommandChoice`.

        Args:
            opt1 (:obj:`str`): Option that this callback will do autocomplete for.
            *opts (:obj:`str`): Additional options that this callback will do autocomplete for.

        Example:

            .. code-block:: python

                @lightbulb.option("foo", "bar", autocomplete=True)
                @lightbulb.command(...)
                @lightbulb.implements(lightbulb.SlashCommand)
                async def foo(ctx: lightbulb.Context) -> None
                    ...

                @foo.autocomplete("foo")  # Name of the option you want to autocomplete
                async def foo_autocomplete(
                    opt: hikari.AutocompleteInteractionOption, inter: hikari.AutocompleteInteraction
                ) -> Union[str, Sequence[str], hikari.api.AutocompleteChoiceBuilder, Sequence[hikari.api.AutocompleteChoiceBuilder]]:
                    ...
        """

        def decorate(func: AutocompleteCallbackT) -> AutocompleteCallbackT:
            for opt in [opt1, *opts]:
                self._autocomplete_callbacks[opt] = func
            return func

        return decorate


class SubCommandTrait(abc.ABC):
    """
    Trait that all subcommands and subgroups have.

    You can check if any given command is a subcommand by checking ``issubclass``
    on the command's class or ``isinstance`` if you have the object.
    """

    __slots__ = ()


class Command(abc.ABC):
    """
    Abstract base class for all command types.

    Args:
        app (:obj:`~.app.BotApp`): The ``BotApp`` instance that the command is registered to.
        initialiser (:obj:`~CommandLike`): The ``CommandLike`` object to create the command from.
    """

    __slots__ = (
        "_initialiser",
        "app",
        "parent",
        "_plugin",
        "_max_concurrency_semaphores",
    )

    def __init__(self, app: app_.BotApp, initialiser: CommandLike) -> None:
        self._initialiser = initialiser
        self._plugin: t.Optional[plugins.Plugin] = None
        self.app: app_.BotApp = app
        """The ``BotApp`` instance the command is registered to."""
        self.parent: t.Optional[Command] = None
        """The parent for the command."""
        self._max_concurrency_semaphores: t.Dict[t.Hashable, asyncio.Semaphore] = {}

    @property
    def _help_getter(self) -> t.Optional[t.Callable[[Command, context_.base.Context], str]]:
        return self._initialiser.help_getter

    @property
    def callback(self) -> CommandCallbackT:
        """The callback function for the command."""
        return self._initialiser.callback

    @property
    def name(self) -> str:
        """The name of the command."""
        return self._initialiser.name

    @property
    def description(self) -> str:
        """The description of the command."""
        return self._initialiser.description

    @property
    def options(self) -> t.MutableMapping[str, OptionLike]:
        """The options for the command."""
        return self._initialiser.options

    @property
    def checks(self) -> t.Sequence[t.Union[checks.Check, checks._ExclusiveCheck]]:
        """The checks for the command."""
        return self._initialiser.checks

    @property
    def aliases(self) -> t.Sequence[str]:
        """The aliases for the command. This value means nothing for application commands."""
        return self._initialiser.aliases

    @property
    def error_handler(
        self,
    ) -> t.Optional[t.Callable[[events.CommandErrorEvent], t.Coroutine[t.Any, t.Any, t.Optional[bool]]]]:
        """The error handler function for the command."""
        return self._initialiser.error_handler

    @property
    def parser(self) -> t.Optional[t.Type[parser_.BaseParser]]:
        """The argument parser to use for prefix commands."""
        return self._initialiser.parser

    @property
    def cooldown_manager(self) -> t.Optional[cooldowns.CooldownManager]:
        """The cooldown manager instance to use for the command."""
        return self._initialiser.cooldown_manager

    @property
    def auto_defer(self) -> bool:
        """Whether to automatically defer the response when the command is invoked."""
        return self._initialiser.auto_defer

    @property
    def default_ephemeral(self) -> bool:
        """Whether to send responses from this command as ephemeral messages by default."""
        return self._initialiser.ephemeral

    @property
    def check_exempt(self) -> t.Callable[[context_.base.Context], t.Union[bool, t.Coroutine[t.Any, t.Any, bool]]]:
        """Check exempt predicate to use for the command."""
        return self._initialiser.check_exempt or (lambda _: False)

    @property
    def hidden(self) -> bool:
        """Whether the command should be hidden from the help command."""
        return self._initialiser.hidden

    @property
    def inherit_checks(self) -> bool:
        """Whether the command should inherit checks from the parent group."""
        return self._initialiser.inherit_checks

    @property
    def pass_options(self) -> bool:
        """Whether the command will have its options passed as keyword arguments when invoked."""
        return self._initialiser.pass_options

    @property
    def max_concurrency(self) -> t.Optional[t.Tuple[int, t.Type[buckets.Bucket]]]:
        """The max concurrency rule for the command."""
        return self._initialiser.max_concurrency

    @property
    def app_command_default_member_permissions(self) -> t.Optional[hikari.Permissions]:
        """The default member permissions for this command, as an application command."""
        return self._initialiser.app_command_default_member_permissions

    @property
    def app_command_dm_enabled(self) -> bool:
        """Whether this command, as an application command, is enabled in DMs."""
        return self._initialiser.app_command_dm_enabled

    @property
    def app_command_bypass_author_permission_checks(self) -> bool:
        """Whether invocations of this command as an application command will bypass author permission checks."""
        return self._initialiser.app_command_bypass_author_permission_checks

    @property
    def name_localizations(self) -> t.Mapping[t.Union[hikari.Locale, str], str]:
        """
        A mapping of locale to name localizations for this command

        .. versionadded:: 2.3.0
        """
        return self._initialiser.name_localizations

    @property
    def description_localizations(self) -> t.Mapping[t.Union[hikari.Locale, str], str]:
        """
        A mapping of locale to description localizations for this command

        .. versionadded:: 2.3.0
        """
        return self._initialiser.description_localizations

    @property
    def nsfw(self) -> bool:
        """
        Whether the command should only be enabled in NSFW channels.

        For prefix commands, this will add an NSFW-channel only check to the command automatically.
        For slash commands, this will behave as specified in the Discord documentation.

        .. versionadded:: 2.3.1
        """
        return self._initialiser.nsfw

    def __hash__(self) -> int:
        return hash(self.name)

    async def __call__(self, context: context_.base.Context, **kwargs: t.Any) -> None:
        if self.pass_options:
            for opt in context.raw_options:
                kwargs.setdefault(opt, context.raw_options[opt])
        return await self.callback(context, **kwargs)

    def _validate_attributes(self) -> None:
        pass

    def _set_plugin(self, pl: plugins.Plugin) -> None:
        self._plugin = pl

    @property
    def plugin(self) -> t.Optional[plugins.Plugin]:
        """The plugin that the command belongs to."""
        return self._plugin

    @plugin.setter
    def plugin(self, pl: plugins.Plugin) -> None:
        self._set_plugin(pl)

    @property
    def bot(self) -> app_.BotApp:
        """Alias for :obj:`~Context.app`"""
        return self.app

    def get_help(self, context: context_.base.Context) -> str:
        """
        Get the help text for the command under the given context. This method calls the help getter
        provided by the :obj:`~.decorators.set_help` decorator. An empty string will be returned
        if no help getter function was set.

        Args:
            context (:obj:`~.context.base.Context`): Context to get the help text under.

        Returns:
            :obj:`str`: Command's help text.
        """
        if self._help_getter is None:
            return ""
        return self._help_getter(self, context)

    @property
    def is_subcommand(self) -> bool:
        """Boolean representing whether this object is a subcommand."""
        return isinstance(self, SubCommandTrait)

    @property
    def qualname(self) -> str:
        """The qualified name for the command."""
        return self.name

    @property
    @abc.abstractmethod
    def signature(self) -> str:
        """The command's text signature."""
        ...

    async def _evaluate_max_concurrency(self, context: context_.base.Context) -> None:
        if self.max_concurrency is None:
            return
        bucket_hash = self.max_concurrency[1].extract_hash(context)
        if bucket_hash not in self._max_concurrency_semaphores:
            self._max_concurrency_semaphores[bucket_hash] = asyncio.Semaphore(self.max_concurrency[0])
        if self._max_concurrency_semaphores[bucket_hash].locked():
            assert context.invoked is not None
            raise errors.MaxConcurrencyLimitReached(
                f"Maximum concurrency limit for command '{context.invoked.qualname}' exceeded",
                bucket=self.max_concurrency[1],
            )
        await self._max_concurrency_semaphores[bucket_hash].acquire()

    def _release_max_concurrency(self, context: context_.base.Context) -> None:
        if self.max_concurrency is None:
            return

        if sem := self._max_concurrency_semaphores.get(self.max_concurrency[1].extract_hash(context)):
            sem.release()

    async def invoke(self, context: context_.base.Context, **kwargs: t.Any) -> None:
        """
        Invokes the command under the given context. All checks, cooldowns and concurrency limits will be processed
        prior to invocation.
        """
        context._invoked = self

        await self._evaluate_max_concurrency(context)
        try:
            await self.evaluate_checks(context)
            await self.evaluate_cooldowns(context)
            await self(context, **kwargs)
        except Exception:
            raise
        finally:
            self._release_max_concurrency(context)

    async def evaluate_checks(self, context: context_.base.Context) -> bool:
        """
        Evaluate the command's checks under the given context. This method will either return
        ``True`` if all the checks passed, or it will raise :obj:`~.errors.CheckFailure`.
        """
        exempt = self.check_exempt(context)
        if inspect.iscoroutine(exempt):
            exempt = await exempt
        if exempt:
            return True

        parent_checks = self.parent.checks if self.inherit_checks and self.parent is not None else []

        failed_checks: t.List[errors.CheckFailure] = []
        for check in [*self.app._checks, *getattr(self.plugin, "_checks", []), *self.checks, *parent_checks]:
            try:
                result = check(context)
                if inspect.iscoroutine(result):
                    result = await result

                if not result:
                    failed_checks.append(errors.CheckFailure(f"Check {check.__name__} failed for command {self.name}"))
            except Exception as ex:
                if not isinstance(ex, errors.CheckFailure):
                    error = errors.CheckFailure(str(ex))
                    error.__cause__ = ex
                else:
                    error = ex
                failed_checks.append(error)

        if len(failed_checks) > 1:
            raise errors.CheckFailure(
                "Multiple checks failed: " + ", ".join(str(ex) for ex in failed_checks), causes=failed_checks
            )
        elif failed_checks:
            raise failed_checks[0]

        return True

    async def evaluate_cooldowns(self, context: context_.base.Context) -> None:
        """
        Evaluate the command's cooldown under the given context. This method will either return
        ``None`` if the command is not on cooldown or raise :obj:`.errors.CommandIsOnCooldown`.
        """
        if self.cooldown_manager is not None:
            await self.cooldown_manager.add_cooldown(context)


class ApplicationCommand(Command, abc.ABC):
    """Abstract base class for all application command types."""

    __slots__ = ("_guilds", "instances")

    def __init__(self, app: app_.BotApp, initialiser: CommandLike) -> None:
        super().__init__(app, initialiser)
        self._guilds = initialiser.guilds
        self.instances: t.Dict[t.Union[int, None], hikari.PartialCommand] = {}
        """Mapping of guild ID to created hikari ``PartialCommand`` objects for this command."""

    @property
    def guilds(self) -> t.Sequence[int]:
        """The guilds that this command is available in."""
        return self.app.default_enabled_guilds if self._guilds is hikari.UNDEFINED else self._guilds

    @property
    def signature(self) -> str:
        sig = self.qualname
        if self.options:
            sig += f" {' '.join(f'<{o.name}>' if o.required else f'[{o.name}={o.default}]' for o in self.options.values())}"
        return sig

    async def create(self, guild: t.Optional[int] = None) -> hikari.PartialCommand:
        """
        Creates the command in the guild with the given ID, or globally if no
        guild ID was provided.

        Args:
            guild (Optional[:obj:`int`]): ID of the guild to create the command in, or ``None`` if to create globally.

        Returns:
            :obj:`~hikari.commands.PartialCommand`: Created hikari ``Command`` object.
        """
        assert self.app.application is not None
        kwargs = self.as_create_kwargs()
        kwargs.update({"guild": guild} if guild is not None else {})
        kwargs["nsfw"] = self.nsfw

        if guild is None:
            kwargs["dm_enabled"] = self.app_command_dm_enabled

        if self.app_command_default_member_permissions is not None:
            kwargs["default_member_permissions"] = self.app_command_default_member_permissions

        if self.name_localizations:
            kwargs["name_localizations"] = self.name_localizations

        if self.description_localizations:
            kwargs["description_localizations"] = self.description_localizations

        cmd_type: hikari.CommandType = kwargs.pop("type")
        created_cmd: hikari.PartialCommand
        if cmd_type is hikari.CommandType.SLASH:
            created_cmd = await self.app.rest.create_slash_command(
                self.app.application,
                **kwargs,
            )
        else:
            created_cmd = await self.app.rest.create_context_menu_command(
                self.app.application,
                type=cmd_type,
                **kwargs,
            )

        self.instances[guild] = created_cmd
        assert isinstance(created_cmd, hikari.PartialCommand)
        return created_cmd

    async def _auto_create(self) -> None:
        if self.guilds:
            for guild_id in self.guilds:
                await self.create(guild_id)
        else:
            await self.create()

    async def delete(self, guild: t.Optional[int]) -> None:
        """
        Deletes the command in the guild with the given ID, or globally if no
        guild ID was provided.

        Args:
            guild (Optional[:obj:`int`]): ID of the guild to delete the command in, or ``None`` if to delete globally.

        Returns:
            ``None``
        """
        cmd = self.instances.pop(guild, None)
        if cmd is None:
            return
        await cmd.delete()

    async def _auto_delete(self) -> None:
        for cmd in self.instances.values():
            await cmd.delete()
        self.instances.clear()

    @abc.abstractmethod
    def as_create_kwargs(self) -> t.Dict[str, t.Any]:
        """
        Converts this class into a dictionary of kwargs required by
        :obj:`~hikari.api.rest.RESTClient.create_application_command`.

        Returns:
            Dict[:obj:`str`, Any]: Kwargs required in order to create the command.
        """
        ...



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = [
    "BooleanConverter",
    "UserConverter",
    "MemberConverter",
    "GuildChannelConverter",
    "TextableGuildChannelConverter",
    "GuildCategoryConverter",
    "GuildVoiceChannelConverter",
    "RoleConverter",
    "EmojiConverter",
    "GuildConverter",
    "MessageConverter",
    "InviteConverter",
    "ColourConverter",
    "ColorConverter",
    "TimestampConverter",
    "SnowflakeConverter",
]

import datetime
import re
import typing as t

import hikari

from lightbulb.converters import base
from lightbulb.utils import search

if t.TYPE_CHECKING:
    from lightbulb import context as context_

T = t.TypeVar("T")

USER_MENTION_REGEX: t.Final[re.Pattern[str]] = re.compile(r"<@!?(\d+)>")
CHANNEL_MENTION_REGEX: t.Final[re.Pattern[str]] = re.compile(r"<#(\d+)>")
ROLE_MENTION_REGEX: t.Final[re.Pattern[str]] = re.compile(r"<@&(\d+)>")
EMOJI_MENTION_REGEX: t.Final[re.Pattern[str]] = re.compile(r"<a?:\w+:(\d+)>")
TIMESTAMP_MENTION_REGEX: t.Final[re.Pattern[str]] = re.compile(r"<t:(\d+)(?::[tTdDfFR])?>")
CHANNEL_MESSAGE_REGEX: t.Final[re.Pattern[str]] = re.compile(r"(\d+)-(\d+)")

BOOLEAN_MAPPING: t.Dict[str, bool] = {"yes": True, "y": True, "1": True, "no": False, "n": False, "0": False}


def _resolve_id_from_arg(arg_string: str, regex: re.Pattern[str]) -> hikari.Snowflake:
    if match := regex.match(arg_string):
        arg_string = match.group(1)
    return hikari.Snowflake(arg_string)


async def _get_or_fetch_guild_channel_from_id(
    context: context_.base.Context, channel_id: hikari.Snowflake
) -> t.Optional[hikari.GuildChannel]:
    channel = context.app.cache.get_guild_channel(channel_id)
    if channel is None:
        channel = await context.app.rest.fetch_channel(channel_id)  # type: ignore
    return channel


async def _try_convert_to_guild_channel(context: context_.base.Context, arg: str) -> t.Optional[hikari.GuildChannel]:
    if context.guild_id is None:
        raise TypeError("Cannot resolve a guild channel object for a command run outside of a guild")

    channel: t.Optional[hikari.GuildChannel]
    try:
        channel_id = _resolve_id_from_arg(arg, CHANNEL_MENTION_REGEX)
    except ValueError:
        channels = context.app.cache.get_guild_channels_view_for_guild(context.guild_id)
        channel = search.get(channels.values(), name=arg)
    else:
        channel = await _get_or_fetch_guild_channel_from_id(context, channel_id)
    return channel


def _raise_if_not_none(obj: t.Optional[T]) -> T:
    if obj is None:
        raise TypeError("No object could be resolved from the argument")
    return obj


class BooleanConverter(base.BaseConverter[bool]):
    """Implementation of the base converter for converting arguments into a boolean."""

    __slots__ = ()

    async def convert(self, arg: str) -> bool:
        try:
            return BOOLEAN_MAPPING[arg]
        except KeyError:
            raise TypeError("Invalid input for boolean type. Valid inputs are: 'yes', 'y', '1', 'no', 'n', '0'")


class UserConverter(base.BaseConverter[hikari.User]):
    """Implementation of the base converter for converting arguments into a User object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.User:
        try:
            user_id = _resolve_id_from_arg(arg, USER_MENTION_REGEX)
        except ValueError:
            users = self.context.app.cache.get_users_view()
            user = search.find(users.values(), lambda u: u.username == arg or f"{u.username}#{u.discriminator}" == arg)
        else:
            user = self.context.app.cache.get_user(user_id)
            if user is None:
                user = await self.context.app.rest.fetch_user(user_id)
        return _raise_if_not_none(user)


class MemberConverter(base.BaseConverter[hikari.Member]):
    """Implementation of the base converter for converting arguments into a Member object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.Member:
        if self.context.guild_id is None:
            raise TypeError("Cannot resolve a member object for a command run outside of a guild")
        try:
            user_id = _resolve_id_from_arg(arg, USER_MENTION_REGEX)
        except ValueError:
            members = self.context.app.cache.get_members_view_for_guild(self.context.guild_id)
            member = search.find(
                members.values(),
                lambda m: m.username == arg or m.nickname == arg or f"{m.username}#{m.discriminator}" == arg,
            )
        else:
            member = self.context.app.cache.get_member(self.context.guild_id, user_id)
            if member is None:
                member = await self.context.app.rest.fetch_member(self.context.guild_id, user_id)
        return _raise_if_not_none(member)


class GuildChannelConverter(base.BaseConverter[hikari.GuildChannel]):
    """Implementation of the base converter for converting arguments into a GuildChannel object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.GuildChannel:
        channel = await _try_convert_to_guild_channel(self.context, arg)
        if not isinstance(channel, hikari.GuildChannel):
            raise TypeError("No object could be resolved from the argument")
        return channel


class TextableGuildChannelConverter(base.BaseConverter[hikari.TextableGuildChannel]):
    """Implementation of the base converter for converting arguments into a TextableGuildChannel object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.TextableGuildChannel:
        channel = await _try_convert_to_guild_channel(self.context, arg)
        if not isinstance(channel, hikari.TextableGuildChannel):
            raise TypeError("No object could be resolved from the argument")
        return channel


class GuildCategoryConverter(base.BaseConverter[hikari.GuildCategory]):
    """Implementation of the base converter for converting arguments into a GuildCategory object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.GuildCategory:
        channel = await _try_convert_to_guild_channel(self.context, arg)
        if not isinstance(channel, hikari.GuildCategory):
            raise TypeError("No object could be resolved from the argument")
        return channel


class GuildVoiceChannelConverter(base.BaseConverter[hikari.GuildVoiceChannel]):
    """Implementation of the base converter for converting arguments into a GuildVoiceChannel object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.GuildVoiceChannel:
        channel = await _try_convert_to_guild_channel(self.context, arg)
        if not isinstance(channel, hikari.GuildVoiceChannel):
            raise TypeError("No object could be resolved from the argument")
        return channel


class RoleConverter(base.BaseConverter[hikari.Role]):
    """Implementation of the base converter for converting arguments into a Role object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.Role:
        if self.context.guild_id is None:
            raise TypeError("Cannot resolve a role object for a command run outside of a guild")
        try:
            role_id = _resolve_id_from_arg(arg, ROLE_MENTION_REGEX)
        except ValueError:
            roles = self.context.app.cache.get_roles_view_for_guild(self.context.guild_id)
            role = search.get(roles.values(), name=arg)
        else:
            role = self.context.app.cache.get_role(role_id)
            if role is None:
                fetched_roles = await self.context.app.rest.fetch_roles(self.context.guild_id)
                role = {r.id: r for r in fetched_roles}.get(role_id)
        return _raise_if_not_none(role)


class EmojiConverter(base.BaseConverter[hikari.Emoji]):
    """Implementation of the base converter for converting arguments into an Emoji object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.Emoji:
        return hikari.Emoji.parse(arg)


class GuildConverter(base.BaseConverter[hikari.GuildPreview]):
    """Implementation of the base converter for converting arguments into a GuildPreview object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.GuildPreview:
        if self.context.guild_id is None:
            raise TypeError("Cannot resolve a guild object for a command run outside of a guild")
        guild_preview = None
        if arg.isdigit():
            guild_id = int(arg)
            guild_preview = await self.context.app.rest.fetch_guild_preview(guild_id)
        else:
            guilds = self.context.app.cache.get_available_guilds_view()
            guild = search.get(guilds.values(), name=arg)
            if guild is not None:
                guild_preview = await self.context.app.rest.fetch_guild_preview(guild.id)
        return _raise_if_not_none(guild_preview)


class MessageConverter(base.BaseConverter[hikari.Message]):
    """Implementation of the base converter for converting arguments into a Message object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.Message:
        if arg.isdigit():
            m_id, c_id = int(arg), int(self.context.channel_id)
        elif match := CHANNEL_MESSAGE_REGEX.match(arg):
            m_id, c_id = int(match.group(2)), int(match.group(1))
        else:
            parts = arg.rstrip("/").split("/")
            m_id, c_id = int(parts[-1]), int(parts[-2])

        if c_id != self.context.channel_id:
            raise ValueError("Cannot convert to message for a different channel than the command was invoked in")

        return await self.context.app.rest.fetch_message(c_id, m_id)


class InviteConverter(base.BaseConverter[hikari.Invite]):
    """Implementation of the base converter for converting arguments into an Invite object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.Invite:
        inv_code = arg.rstrip("/").split("/")[-1]
        invite: t.Optional[hikari.Invite] = self.context.app.cache.get_invite(inv_code)
        if invite is None:
            invite = await self.context.app.rest.fetch_invite(inv_code)
        return _raise_if_not_none(invite)


class ColourConverter(base.BaseConverter[hikari.Colour]):
    """Implementation of the base converter for converting arguments into a Colour object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.Colour:
        return hikari.Colour.of(arg)


ColorConverter = ColourConverter
"""Alias for :obj:`~ColorConverter`."""


class TimestampConverter(base.BaseConverter[datetime.datetime]):
    """Implementation of the base converter for converting arguments into a datetime object."""

    __slots__ = ()

    async def convert(self, arg: str) -> datetime.datetime:
        timestamp: t.Optional[str] = None
        if match := TIMESTAMP_MENTION_REGEX.match(arg):
            timestamp = match.group(1)
        if timestamp is None:
            raise TypeError("Could not resolve timestamp")
        return datetime.datetime.fromtimestamp(int(timestamp), datetime.timezone.utc)


class SnowflakeConverter(base.BaseConverter[hikari.Snowflake]):
    """Implementation of the base converter for converting arguments into a Snowflake object."""

    __slots__ = ()

    async def convert(self, arg: str) -> hikari.Snowflake:
        try:
            snowflake = hikari.Snowflake(arg.strip("<#!&>"))
        except ValueError:
            raise TypeError("Could not resolve snowflake")
        return snowflake



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["CooldownManager"]

import inspect
import time
import typing as t

from lightbulb import buckets
from lightbulb import cooldown_algorithms
from lightbulb import errors

if t.TYPE_CHECKING:
    from lightbulb.context import base as ctx_base


class CooldownManager:
    """
    The cooldown manager for a command.
    """

    __slots__ = ("callback", "cooldowns")

    def __init__(
        self,
        callback: t.Callable[[ctx_base.Context], t.Union[buckets.Bucket, t.Coroutine[t.Any, t.Any, buckets.Bucket]]],
    ) -> None:
        self.callback = callback
        self.cooldowns: t.MutableMapping[t.Hashable, buckets.Bucket] = {}
        """Mapping of a hashable to a :obj:`~.buckets.Bucket` representing the currently stored cooldowns."""

    async def _get_bucket(self, context: ctx_base.Context) -> buckets.Bucket:
        bucket = self.callback(context)
        if inspect.iscoroutine(bucket):
            bucket = await bucket
        assert isinstance(bucket, buckets.Bucket)
        return bucket

    async def add_cooldown(self, context: ctx_base.Context) -> None:
        """
        Add a cooldown under the given context. If an expired bucket already exists then it
        will be overwritten.

        Args:
            context (:obj:`~.context.base.Context`): The context to add a cooldown under.

        Returns:
            ``None``

        Raises:
            :obj:`~.errors.CommandIsOnCooldown`: The command is currently on cooldown for the given context.
        """
        bucket = await self._get_bucket(context)

        cooldown_hash = bucket.extract_hash(context)
        cooldown_bucket = self.cooldowns.get(cooldown_hash)

        if cooldown_bucket is not None:
            cooldown_status = cooldown_bucket.acquire()
            if inspect.iscoroutine(cooldown_status):
                cooldown_status = await cooldown_status

            if cooldown_status is cooldown_algorithms.CooldownStatus.ACTIVE:
                # Cooldown has been activated
                assert cooldown_bucket.start_time is not None
                raise errors.CommandIsOnCooldown(
                    "This command is on cooldown",
                    retry_after=(cooldown_bucket.start_time + cooldown_bucket.length) - time.perf_counter(),
                )
            elif cooldown_status is cooldown_algorithms.CooldownStatus.INACTIVE:
                # Cooldown has not yet been activated.
                return

        self.cooldowns[cooldown_hash] = bucket
        maybe_coro = self.cooldowns[cooldown_hash].acquire()
        if inspect.iscoroutine(maybe_coro):
            await maybe_coro

    async def reset_cooldown(self, context: ctx_base.Context) -> None:
        """
        Reset the cooldown under the given context.

        Args:
            context (:obj:`~.context.base.Context`): The context to reset the cooldown under.

        Returns:
            ``None``
        """
        bucket = await self._get_bucket(context)
        del self.cooldowns[bucket.extract_hash(context)]



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["SlashContext"]

import asyncio
import inspect
import typing as t

import hikari

from lightbulb import commands
from lightbulb.context import base
from lightbulb.converters import CONVERTER_TYPE_MAPPING
from lightbulb.converters import BaseConverter

if t.TYPE_CHECKING:
    from lightbulb import app as app_


class SlashContext(base.ApplicationContext):
    """
    An implementation of :obj:`~.context.base.Context` for slash commands.

    Args:
        app (:obj:`~.app.BotApp`): The ``BotApp`` instance that the context is linked to.
        event (:obj:`~hikari.events.interaction_events.InteractionCreateEvent`): The event to create the context
            from.
        command (:obj:`~.commands.slash.SlashCommand`): The command that the context is for.
    """

    __slots__ = ("_options", "_raw_options", "_to_convert")

    def __init__(
        self, app: app_.BotApp, event: hikari.InteractionCreateEvent, command: commands.slash.SlashCommand
    ) -> None:
        super().__init__(app, event, command)
        self._options: t.Dict[str, t.Any] = {}
        self._raw_options: t.Sequence[hikari.CommandInteractionOption] = self.interaction.options or []
        self._to_convert: t.List[t.Coroutine[t.Any, t.Any, None]] = []
        self._parse_options(self.interaction.options)

    async def _convert_option(self, name: str, value: str) -> None:
        cmd = self.invoked or self.command
        opt_type = cmd.options[name].arg_type
        if opt_type in CONVERTER_TYPE_MAPPING:
            self._options[name] = await CONVERTER_TYPE_MAPPING[cmd.options[name].arg_type](self).convert(value)
        elif inspect.isclass(opt_type) and issubclass(cmd.options[name].arg_type, BaseConverter):
            self._options[name] = await opt_type.convert(value)
        elif callable(opt_type):
            temp: t.Any = opt_type(value)
            if inspect.iscoroutine(temp):
                temp = await temp
            self._options[name] = temp
        else:
            self._options[name] = value

    def _parse_options(self, options: t.Optional[t.Sequence[hikari.CommandInteractionOption]]) -> None:
        # We need to clear the options here to ensure the subcommand name does not exist in the mapping
        self._options.clear()
        for opt in options or []:
            # Why is mypy so annoying about this ??
            if opt.type is hikari.OptionType.USER and self.resolved is not None:
                val = t.cast(hikari.Snowflake, opt.value)
                self._options[opt.name] = self.resolved.members.get(val, self.resolved.users.get(val, opt.value))
            elif opt.type is hikari.OptionType.CHANNEL and self.resolved is not None:
                val = t.cast(hikari.Snowflake, opt.value)
                self._options[opt.name] = self.resolved.channels.get(val, opt.value)
            elif opt.type is hikari.OptionType.ROLE and self.resolved is not None:
                val = t.cast(hikari.Snowflake, opt.value)
                self._options[opt.name] = self.resolved.roles.get(val, opt.value)
            elif opt.type is hikari.OptionType.ATTACHMENT and self.resolved is not None:
                val = t.cast(hikari.Snowflake, int(opt.value) if isinstance(opt.value, str) else opt.value)
                self._options[opt.name] = self.resolved.attachments.get(val, opt.value)
            else:
                if isinstance(opt.value, str):
                    self._to_convert.append(self._convert_option(opt.name, opt.value))
                    continue
                self._options[opt.name] = opt.value

        cmd = self.invoked or self.command
        for opt in cmd.options.values():
            self._options.setdefault(opt.name, opt.default if opt.default is not hikari.UNDEFINED else None)

    async def _maybe_defer(self) -> None:
        await super()._maybe_defer()
        # Ensure that running converters don't block the automatic deferral
        await asyncio.gather(*self._to_convert)

    @property
    def raw_options(self) -> t.Dict[str, t.Any]:
        return self._options

    @property
    def prefix(self) -> str:
        return "/"

    @property
    def command(self) -> commands.slash.SlashCommand:
        assert isinstance(self._command, commands.slash.SlashCommand)
        return self._command



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["Bucket", "UserBucket", "GuildBucket", "GlobalBucket", "ChannelBucket"]

import abc
import time
import typing as t

from lightbulb import cooldown_algorithms

if t.TYPE_CHECKING:
    from lightbulb.context import base as ctx_base


class Bucket(abc.ABC):
    """
    Base class that represents a bucket that cooldowns or max concurrency limits can
    be applied under. All buckets must inherit from this class.

    Note that for max concurrency limiting, the bucket object will never be instantiated.

    Args:
        length (:obj:`float`): Length of the cooldown timer.
        max_usages (:obj:`int`): Number of command usages before the cooldown is activated.
    """

    __slots__ = ("length", "usages", "cooldown_algorithm", "activated", "start_time")

    def __init__(
        self,
        length: float,
        max_usages: int,
        cooldown_algorithm: t.Type[
            cooldown_algorithms.CooldownAlgorithm
        ] = cooldown_algorithms.BangBangCooldownAlgorithm,
    ) -> None:
        self.length = length
        self.usages = max_usages
        self.cooldown_algorithm = cooldown_algorithm()
        self.activated = False
        self.start_time: t.Optional[float] = None
        """The start time of the bucket cooldown. This is relative to :meth:`time.perf_counter`."""

    @classmethod
    @abc.abstractmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        """
        Extracts the hash from the context which links a command usage to a single bucket.

        Args:
            context (:obj:`~.context.base.Context`): The context the command was invoked under.

        Returns:
            Hashable: Hashable object linking the context to a bucket.
        """
        ...

    def acquire(
        self,
    ) -> t.Union[cooldown_algorithms.CooldownStatus, t.Coroutine[t.Any, t.Any, cooldown_algorithms.CooldownStatus]]:
        """
        Get the current state of the cooldown and add a command usage if the cooldown is not
        currently active or has expired.

        This method is **only** used when processing cooldowns.

        Returns:
            :obj:`~CooldownStatus`: The status of the cooldown bucket.
        """
        if self.active:
            return cooldown_algorithms.CooldownStatus.ACTIVE
        elif self.activated and self.expired:
            return cooldown_algorithms.CooldownStatus.EXPIRED

        return self.cooldown_algorithm.evaluate(self)

    @property
    def active(self) -> bool:
        """
        Whether the cooldown represented by the bucket is currently active.
        """
        return self.activated and not self.expired

    @property
    def expired(self) -> bool:
        """
        Whether the cooldown represented by the bucket has expired.
        """
        if self.start_time is not None:
            return time.perf_counter() >= (self.start_time + self.length)
        return True

    def activate(self) -> None:
        self.activated = True
        self.start_time = time.perf_counter()


class GlobalBucket(Bucket):
    """
    Global bucket. All cooldowns and concurrency limits will
    be applied globally if you use this bucket.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        return 0


class UserBucket(Bucket):
    """
    User bucket. All cooldowns and concurrency limits will
    be applied per user if you use this bucket.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        return context.author.id


class ChannelBucket(Bucket):
    """
    Channel bucket. All cooldowns and concurrency limits will
    be applied per channel if you use this bucket.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        return context.channel_id


class GuildBucket(Bucket):
    """
    Guild bucket. All cooldowns and concurrency limits will be applied per guild
    if you use this bucket. The command will still be permitted to be run in DMs in which
    case the DM channel ID will be used as the cooldown identifier instead of a guild ID.
    """

    __slots__ = ()

    @classmethod
    def extract_hash(cls, context: ctx_base.Context) -> t.Hashable:
        return context.guild_id if context.guild_id is not None else context.channel_id



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = [
    "CooldownStatus",
    "CooldownAlgorithm",
    "BangBangCooldownAlgorithm",
    "FixedWindowCooldownAlgorithm",
    "SlidingWindowCooldownAlgorithm",
]

import abc
import enum
import time
import typing as t

from lightbulb import errors

if t.TYPE_CHECKING:
    from lightbulb import buckets


class CooldownStatus(enum.Enum):
    INACTIVE = enum.auto()
    ACTIVE = enum.auto()
    EXPIRED = enum.auto()


class CooldownAlgorithm(abc.ABC):
    """
    Base class that represents an algorithm that can be used to calculate whether a command
    is on cooldown or not.
    """

    __slots__ = ()

    @abc.abstractmethod
    def evaluate(self, bucket: buckets.Bucket) -> t.Union[CooldownStatus, t.Coroutine[t.Any, t.Any, CooldownStatus]]:
        """
        Evaluate whether the command should be on cooldown given internal state of this object and
        the bucket object passed in.

        Args:
            bucket (:obj:`~.buckets.Bucket`): The bucket to process the cooldown for.

        Returns:
            :obj:`~CooldownStatus`: The status of the cooldown.
        """
        ...


class BangBangCooldownAlgorithm(CooldownAlgorithm):
    """
    Cooldown algorithm that allows ``n`` command invocations to be used, and then waits for the appropriate
    amount of time before allowing any additional invocations.
    """

    __slots__ = ("_commands_run",)

    def __init__(self) -> None:
        self._commands_run: int = 0

    def evaluate(self, bucket: buckets.Bucket) -> CooldownStatus:
        self._commands_run += 1
        if self._commands_run >= bucket.usages:
            bucket.activate()
        return CooldownStatus.INACTIVE


class FixedWindowCooldownAlgorithm(CooldownAlgorithm):
    """
    Cooldown algorithm that allows ``n`` command invocations within a fixed window time period. I.e. 5 invocations
    within a 30-second time period, allowing additional invocations once the entire period has elapsed.
    """

    __slots__ = ("_commands_run", "_expires")

    def __init__(self) -> None:
        self._commands_run: int = 0
        self._expires: t.Optional[float] = None

    def evaluate(self, bucket: buckets.Bucket) -> CooldownStatus:
        self._expires = self._expires or (time.perf_counter() + bucket.length)
        if self._expires > time.perf_counter():
            self._commands_run = 0

        self._commands_run += 1
        if self._commands_run > bucket.usages:
            raise errors.CommandIsOnCooldown(
                "This command is on cooldown", retry_after=(self._expires - time.perf_counter())
            )
        return CooldownStatus.INACTIVE


class SlidingWindowCooldownAlgorithm(CooldownAlgorithm):
    """
    Cooldown algorithm that allows ``n`` command invocations within a sliding window time period. I.e. 5 invocations
    within the last 30 seconds will allow an extra invocation every time the earliest invocation falls out of the
    window.
    """

    __slots__ = ("_prev_usages",)

    def __init__(self) -> None:
        self._prev_usages: t.List[float] = []

    def evaluate(self, bucket: buckets.Bucket) -> CooldownStatus:
        interval = time.perf_counter() - bucket.length
        # Find the usages that would be within the current window
        usages_in_window = [usage for usage in self._prev_usages[::-1] if usage > interval]
        if len(usages_in_window) + 1 > bucket.usages:
            raise errors.CommandIsOnCooldown(
                "This command is on cooldown", retry_after=bucket.length - (time.perf_counter() - usages_in_window[0])
            )
        # Replace the stored usages so that the list doesn't grow forever
        self._prev_usages = usages_in_window
        self._prev_usages.append(time.perf_counter())
        return CooldownStatus.INACTIVE



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["PrefixContext"]

import asyncio
import typing as t

import hikari

from lightbulb.context import base

if t.TYPE_CHECKING:
    from hikari.api import special_endpoints

    from lightbulb import app as app_
    from lightbulb import commands
    from lightbulb import parser


class PrefixContext(base.Context):
    """
    An implementation of :obj:`~.context.base.Context` for prefix commands.

    Args:
        app (:obj:`~.app.BotApp`): The ``BotApp`` instance that the context is linked to.
        event (:obj:`~hikari.events.message_events.MessageCreateEvent`): The event to create the context from.
        command (Optional[:obj:`~.commands.prefix.PrefixCommand`]): The command that the context is for, or ``None``
            if no command could be resolved.
        invoked_with (:obj:`str`): The name or alias that the command was invoked with.
        prefix (:obj:`str`): The prefix that was used in this context.
    """

    __slots__ = ("_parser", "_event", "_command", "_invoked_with", "_prefix", "_options")

    def __init__(
        self,
        app: app_.BotApp,
        event: hikari.MessageCreateEvent,
        command: t.Optional[commands.prefix.PrefixCommand],
        invoked_with: str,
        prefix: str,
    ) -> None:
        super().__init__(app)
        self._event = event
        self._command = command
        self._invoked_with = invoked_with
        self._prefix = prefix
        self._options: t.Dict[str, t.Any] = {}
        self._parser: parser.BaseParser

    async def _maybe_defer(self) -> None:
        if self._deferred:
            return

        if self._command is not None and (self._invoked or self._command).auto_defer:
            await self.app.rest.trigger_typing(self.channel_id)
            self._deferred = True

    @property
    def event(self) -> hikari.MessageCreateEvent:
        return self._event

    @property
    def raw_options(self) -> t.Dict[str, t.Any]:
        return self._options

    @property
    def channel_id(self) -> hikari.Snowflake:
        return self.event.message.channel_id

    @property
    def guild_id(self) -> t.Optional[hikari.Snowflake]:
        return self.event.message.guild_id

    @property
    def attachments(self) -> t.Sequence[hikari.Attachment]:
        return self.event.message.attachments

    @property
    def member(self) -> t.Optional[hikari.Member]:
        return self.event.message.member

    @property
    def author(self) -> hikari.User:
        return self.event.message.author

    @property
    def invoked_with(self) -> str:
        return self._invoked_with

    @property
    def prefix(self) -> str:
        return self._prefix

    @property
    def command(self) -> t.Optional[commands.prefix.PrefixCommand]:
        return self._command

    def get_channel(self) -> t.Optional[t.Union[hikari.GuildChannel, hikari.Snowflake]]:
        if self.guild_id is not None:
            return self.app.cache.get_guild_channel(self.channel_id) or self.app.cache.get_thread(self.channel_id)
        return self.channel_id

    async def respond(
        self, *args: t.Any, delete_after: t.Union[int, float, None] = None, **kwargs: t.Any
    ) -> base.ResponseProxy:
        """
        Create a response for this context. This method directly calls :obj:`~hikari.messages.Message.respond`. You
        should note that it is not possible to send ephemeral messages as responses to prefix commands. All message flags
        will be removed before the call to :obj:`~hikari.messages.Message.respond`.

        Args:
            *args (Any): Positional arguments passed to :obj:`~hikari.messages.Message.respond`.
            delete_after (Union[int, float, None]): The number of seconds to wait before deleting this response.
            **kwargs (Any): Keyword arguments passed to :obj:`~hikari.messages.Message.respond`.

        Returns:
            :obj:`~hikari.messages.Message`: The created message object.

        .. versionadded:: 2.2.0
            ``delete_after`` kwarg.
        """
        self._deferred = False

        kwargs.pop("flags", None)
        kwargs.pop("response_type", None)

        if args and isinstance(args[0], hikari.ResponseType):
            args = args[1:]

        msg = await self._event.message.respond(*args, **kwargs)
        if delete_after is not None:

            async def _cleanup(timeout: t.Union[int, float]) -> None:
                await asyncio.sleep(timeout)

                try:
                    await msg.delete()
                except hikari.NotFoundError:
                    pass

            self.app.create_task(_cleanup(delete_after))

        self._responses.append(base.ResponseProxy(msg))
        self._responded = True
        return self._responses[-1]

    async def respond_with_modal(
        self,
        title: str,
        custom_id: str,
        component: hikari.UndefinedOr[special_endpoints.ComponentBuilder] = hikari.UNDEFINED,
        components: hikari.UndefinedOr[t.Sequence[special_endpoints.ComponentBuilder]] = hikari.UNDEFINED,
    ) -> t.NoReturn:
        """
        Method only preserved for API consistency. If you attempt to call this method for a prefix command,
        a :obj:`NotImplementedError` will **always** be raised.

        Raises:
            :obj:`NotImplementedError`: **Always**

        .. versionadded:: 2.3.1
        """
        raise NotImplementedError("You cannot respond to a prefix command with a modal")



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["Context", "ApplicationContext", "OptionsProxy", "ResponseProxy"]

import abc
import asyncio
import functools
import typing as t

import hikari

from lightbulb import errors

if t.TYPE_CHECKING:
    from hikari.api import special_endpoints

    from lightbulb import app as app_
    from lightbulb import commands


class OptionsProxy:
    """
    Proxy for the options that the command was invoked with allowing access using
    dot notation as well as dictionary lookup.

    Args:
        options (Dict[:obj:`str`, Any]): Options to act as a proxy for.
    """

    __slots__ = ("_options",)

    def __init__(self, options: t.Dict[str, t.Any]) -> None:
        self._options = options

    def __getattr__(self, item: str) -> t.Any:
        try:
            return self._options[item]
        except KeyError:
            raise AttributeError(f"There is no option called '{item}'") from None

    def __getitem__(self, item: str) -> t.Any:
        try:
            return self._options[item]
        except KeyError:
            raise AttributeError(f"There is no option called '{item}'") from None

    def items(self) -> t.ItemsView[str, t.Any]:
        """
        Iterates through the options and returns a series of key:value
        pairs.

        Returns:
            ItemsView[:obj:`str`, Any]: The options items. This
                is functionally similar to a list of tuples, where for
                each tuple, the key is the option name, and the value is
                the option value.
        """
        return self._options.items()


class ResponseProxy:
    """
    Proxy for context responses. Allows fetching of the message created from the response
    lazily instead of a follow-up request being made immediately.
    """

    __slots__ = ("_message", "_fetcher", "_editor", "_editable", "_deleteable")

    def __init__(
        self,
        message: t.Optional[hikari.Message] = None,
        fetcher: t.Optional[t.Callable[[], t.Coroutine[t.Any, t.Any, hikari.Message]]] = None,
        editor: t.Optional[t.Callable[[ResponseProxy], t.Coroutine[t.Any, t.Any, hikari.Message]]] = None,
        deleteable: bool = True,
    ) -> None:
        if message is None and fetcher is None:
            raise ValueError("One of message or fetcher arguments cannot be None")

        self._message = message
        self._fetcher = fetcher
        self._editor = editor
        self._deleteable = deleteable

        if editor is None:

            async def _default_editor(rp: ResponseProxy, *args: t.Any, **kwargs: t.Any) -> hikari.Message:
                return await (await rp.message()).edit(*args, **kwargs)

            self._editor = _default_editor

    def __await__(self) -> t.Generator[t.Any, None, hikari.Message]:
        return self.message().__await__()

    async def message(self) -> hikari.Message:
        """
        Fetches and/or returns the created message from the context response.

        Returns:
            :obj:`~hikari.messages.Message`: The response's created message.

        Note:
            This object is awaitable (since version `2.2.2`), hence the following is also valid.

            .. code-block:: python

                # Where 'resp' is an instance of ResponseProxy

                # Calling this method
                message = await resp.message()
                # Awaiting the object itself
                message = await resp
        """
        if self._message is not None:
            return self._message
        assert self._fetcher is not None
        msg = await self._fetcher()
        return msg

    async def edit(self, *args: t.Any, **kwargs: t.Any) -> hikari.Message:
        """
        Edits the message that this object is proxying. Shortcut for :obj:`hikari.messages.Message.edit`.

        Args:
            *args: Args passed in to :obj:`hikari.messages.Message.edit`
            **kwargs: Kwargs passed in to :obj:`hikari.messages.Message.edit`

        Returns:
            :obj:`~hikari.messages.Message`: New message after edit.

        Raises:
            :obj:`~.errors.UnsupportedResponseOperation`: This response cannot be edited (for ephemeral
                interaction followup responses).
        """
        assert self._editor is not None
        out = await self._editor(self, *args, **kwargs)
        assert isinstance(out, hikari.Message)
        return out

    async def delete(self) -> None:
        """
        Deletes the message that this object is proxying.

        Returns:
            ``None``

        Raises:
            :obj:`~.errors.UnsupportedResponseOperation`: This response cannot be deleted (for some ephemeral
                interaction responses).
        """
        if not self._deleteable:
            raise errors.UnsupportedResponseOperation("This response does not support deleting.")

        msg = await self.message()
        await msg.delete()


class Context(abc.ABC):
    """
    Abstract base class for all context types.

    Args:
        app (:obj:`~.app.BotApp`): The ``BotApp`` instance that the context is linked to.
    """

    __slots__ = ("_app", "_responses", "_responded", "_deferred", "_invoked")

    def __init__(self, app: app_.BotApp):
        self._app = app
        self._responses: t.List[ResponseProxy] = []
        self._responded: bool = False
        self._deferred: bool = False
        self._invoked: t.Optional[commands.base.Command] = None

    @abc.abstractmethod
    async def _maybe_defer(self) -> None:
        ...

    @property
    def deferred(self) -> bool:
        """Whether the response from this context is currently deferred."""
        return self._deferred

    @property
    def responses(self) -> t.List[ResponseProxy]:
        """List of all previous responses sent for this context."""
        return self._responses

    @property
    def previous_response(self) -> t.Optional[ResponseProxy]:
        """The last response sent for this context."""
        return self._responses[-1] if self._responses else None

    @property
    def interaction(self) -> t.Optional[hikari.CommandInteraction]:
        """The interaction that triggered this context. Will be ``None`` for prefix commands."""
        # Just to keep the interfaces the same for prefix commands and application commands
        return None

    @property
    def resolved(self) -> t.Optional[hikari.ResolvedOptionData]:
        """The resolved option data for this context. Will be ``None`` for prefix commands"""
        # Just to keep the interfaces the same for prefix commands and application commands
        return None

    @property
    def app(self) -> app_.BotApp:
        """The ``BotApp`` instance the context is linked to."""
        return self._app

    @property
    def bot(self) -> app_.BotApp:
        """Alias for :obj:`~Context.app`"""
        return self.app

    @property
    @abc.abstractmethod
    def event(self) -> t.Union[hikari.MessageCreateEvent, hikari.InteractionCreateEvent]:
        """The event for the context."""
        ...

    @property
    def raw_options(self) -> t.Dict[str, t.Any]:
        """Dictionary of :obj:`str` option name to option value that the user invoked the command with."""
        return {}

    @property
    def options(self) -> OptionsProxy:
        """:obj:`~OptionsProxy` wrapping the options that the user invoked the command with."""
        return OptionsProxy(self.raw_options)

    @property
    @abc.abstractmethod
    def channel_id(self) -> hikari.Snowflake:
        """The channel ID for the context."""
        ...

    @property
    @abc.abstractmethod
    def guild_id(self) -> t.Optional[hikari.Snowflake]:
        """The guild ID for the context."""
        ...

    @property
    @abc.abstractmethod
    def attachments(self) -> t.Sequence[hikari.Attachment]:
        ...

    @property
    @abc.abstractmethod
    def member(self) -> t.Optional[hikari.Member]:
        """The member for the context."""
        ...

    @property
    @abc.abstractmethod
    def author(self) -> hikari.User:
        """The author for the context."""
        ...

    @property
    def user(self) -> hikari.User:
        """The user for the context. Alias for :obj:`~Context.author`."""
        return self.author

    @property
    @abc.abstractmethod
    def invoked_with(self) -> str:
        """The command name or alias was used in the context."""
        ...

    @property
    @abc.abstractmethod
    def prefix(self) -> str:
        """The prefix that was used in the context."""

    @property
    @abc.abstractmethod
    def command(self) -> t.Optional[commands.base.Command]:
        """
        The root command object that the context is for.

        See Also:
            :obj:`~Context.invoked`
        """
        ...

    @property
    def invoked(self) -> t.Optional[commands.base.Command]:
        """
        The command or subcommand that was invoked in this context.

        .. versionadded:: 2.1.0
        """
        return self._invoked

    @abc.abstractmethod
    def get_channel(self) -> t.Optional[t.Union[hikari.GuildChannel, hikari.Snowflake]]:
        """The channel object for the context's channel ID."""
        ...

    def get_guild(self) -> t.Optional[hikari.Guild]:
        """The guild object for the context's guild ID."""
        if self.guild_id is None:
            return None
        return self.app.cache.get_guild(self.guild_id)

    async def invoke(self) -> None:
        """
        Invokes the context's command under the current context.

        Returns:
            ``None``
        """
        if self.command is None:
            raise TypeError("This context cannot be invoked - no command was resolved.")
        await self._maybe_defer()
        await self.command.invoke(self)

    @t.overload
    async def respond(
        self,
        response_type: hikari.ResponseType,
        content: hikari.UndefinedOr[t.Any] = hikari.UNDEFINED,
        delete_after: t.Union[int, float, None] = None,
        *,
        attachment: hikari.UndefinedOr[hikari.Resourceish] = hikari.UNDEFINED,
        attachments: hikari.UndefinedOr[t.Sequence[hikari.Resourceish]] = hikari.UNDEFINED,
        component: hikari.UndefinedOr[hikari.api.ComponentBuilder] = hikari.UNDEFINED,
        components: hikari.UndefinedOr[t.Sequence[hikari.api.ComponentBuilder]] = hikari.UNDEFINED,
        embed: hikari.UndefinedOr[hikari.Embed] = hikari.UNDEFINED,
        embeds: hikari.UndefinedOr[t.Sequence[hikari.Embed]] = hikari.UNDEFINED,
        flags: hikari.UndefinedOr[t.Union[int, hikari.MessageFlag]] = hikari.UNDEFINED,
        tts: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        nonce: hikari.UndefinedOr[str] = hikari.UNDEFINED,
        reply: hikari.UndefinedOr[hikari.SnowflakeishOr[hikari.PartialMessage]] = hikari.UNDEFINED,
        reply_must_exist: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        mentions_everyone: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        mentions_reply: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        user_mentions: hikari.UndefinedOr[
            t.Union[hikari.SnowflakeishSequence[hikari.PartialUser], bool]
        ] = hikari.UNDEFINED,
        role_mentions: hikari.UndefinedOr[
            t.Union[hikari.SnowflakeishSequence[hikari.PartialRole], bool]
        ] = hikari.UNDEFINED,
    ) -> ResponseProxy:
        ...

    @t.overload
    async def respond(
        self,
        content: hikari.UndefinedOr[t.Any] = hikari.UNDEFINED,
        delete_after: t.Union[int, float, None] = None,
        *,
        attachment: hikari.UndefinedOr[hikari.Resourceish] = hikari.UNDEFINED,
        attachments: hikari.UndefinedOr[t.Sequence[hikari.Resourceish]] = hikari.UNDEFINED,
        component: hikari.UndefinedOr[hikari.api.ComponentBuilder] = hikari.UNDEFINED,
        components: hikari.UndefinedOr[t.Sequence[hikari.api.ComponentBuilder]] = hikari.UNDEFINED,
        embed: hikari.UndefinedOr[hikari.Embed] = hikari.UNDEFINED,
        embeds: hikari.UndefinedOr[t.Sequence[hikari.Embed]] = hikari.UNDEFINED,
        flags: hikari.UndefinedOr[t.Union[int, hikari.MessageFlag]] = hikari.UNDEFINED,
        tts: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        nonce: hikari.UndefinedOr[str] = hikari.UNDEFINED,
        reply: hikari.UndefinedOr[hikari.SnowflakeishOr[hikari.PartialMessage]] = hikari.UNDEFINED,
        reply_must_exist: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        mentions_everyone: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        mentions_reply: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
        user_mentions: hikari.UndefinedOr[
            t.Union[hikari.SnowflakeishSequence[hikari.PartialUser], bool]
        ] = hikari.UNDEFINED,
        role_mentions: hikari.UndefinedOr[
            t.Union[hikari.SnowflakeishSequence[hikari.PartialRole], bool]
        ] = hikari.UNDEFINED,
    ) -> ResponseProxy:
        ...

    @abc.abstractmethod
    async def respond(
        self, *args: t.Any, delete_after: t.Union[int, float, None] = None, **kwargs: t.Any
    ) -> ResponseProxy:
        """
        Create a response to this context.
        """
        ...

    async def edit_last_response(self, *args: t.Any, **kwargs: t.Any) -> t.Optional[hikari.Message]:
        """
        Edit the most recently sent response. Shortcut for :obj:`hikari.messages.Message.edit`.

        Args:
            *args: Args passed to :obj:`hikari.messages.Message.edit`.
            **kwargs: Kwargs passed to :obj:`hikari.messages.Message.edit`.

        Returns:
            Optional[:obj:`~hikari.messages.Message`]: New message after edit, or ``None`` if no responses have
                been sent for the context yet.
        """
        if not self._responses:
            return None

        return await self._responses[-1].edit(*args, **kwargs)

    async def delete_last_response(self) -> None:
        """
        Delete the most recently send response. Shortcut for :obj:`hikari.messages.Message.delete`.

        Returns:
            ``None``
        """
        if not self._responses:
            return

        await self._responses.pop().delete()

    @abc.abstractmethod
    async def respond_with_modal(
        self,
        title: str,
        custom_id: str,
        component: hikari.UndefinedOr[special_endpoints.ComponentBuilder] = hikari.UNDEFINED,
        components: hikari.UndefinedOr[t.Sequence[special_endpoints.ComponentBuilder]] = hikari.UNDEFINED,
    ) -> None:
        """
        Create a modal response to this context.

        .. versionadded:: 2.3.1
        """
        ...


class ApplicationContext(Context, abc.ABC):
    __slots__ = ("_event", "_interaction", "_command")

    def __init__(
        self, app: app_.BotApp, event: hikari.InteractionCreateEvent, command: commands.base.ApplicationCommand
    ) -> None:
        super().__init__(app)
        self._event = event
        assert isinstance(event.interaction, hikari.CommandInteraction)
        self._interaction: hikari.CommandInteraction = event.interaction
        self._command = command

    async def _maybe_defer(self) -> None:
        if self._deferred:
            return

        if (self._invoked or self._command).auto_defer:
            await self.respond(hikari.ResponseType.DEFERRED_MESSAGE_CREATE)

    @property
    @abc.abstractmethod
    def command(self) -> commands.base.ApplicationCommand:
        ...

    @property
    def event(self) -> hikari.InteractionCreateEvent:
        return self._event

    @property
    def interaction(self) -> hikari.CommandInteraction:
        return self._interaction

    @property
    def channel_id(self) -> hikari.Snowflake:
        return self._interaction.channel_id

    @property
    def guild_id(self) -> t.Optional[hikari.Snowflake]:
        return self._interaction.guild_id

    @property
    def attachments(self) -> t.Sequence[hikari.Attachment]:
        return []

    @property
    def member(self) -> t.Optional[hikari.Member]:
        return self._interaction.member

    @property
    def author(self) -> hikari.User:
        return self._interaction.user

    @property
    def invoked_with(self) -> str:
        return self._command.name

    @property
    def command_id(self) -> hikari.Snowflake:
        return self._interaction.command_id

    @property
    def resolved(self) -> t.Optional[hikari.ResolvedOptionData]:
        return self._interaction.resolved

    def get_channel(self) -> t.Optional[t.Union[hikari.GuildChannel, hikari.Snowflake]]:
        if self.guild_id is not None:
            return self.app.cache.get_guild_channel(self.channel_id) or self.app.cache.get_thread(self.channel_id)
        return self.channel_id

    async def respond(
        self, *args: t.Any, delete_after: t.Union[int, float, None] = None, **kwargs: t.Any
    ) -> ResponseProxy:
        """
        Create a response for this context. The first time this method is called, the initial
        interaction response will be created by calling
        :obj:`~hikari.interactions.command_interactions.CommandInteraction.create_initial_response` with the response
        type set to :obj:`~hikari.interactions.base_interactions.ResponseType.MESSAGE_CREATE` if not otherwise
        specified.

        Subsequent calls will instead create followup responses to the interaction by calling
        :obj:`~hikari.interactions.command_interactions.CommandInteraction.execute`.

        Args:
            *args (Any): Positional arguments passed to ``CommandInteraction.create_initial_response`` or
                ``CommandInteraction.execute``.
            delete_after (Union[:obj:`int`, :obj:`float`, ``None``]): The number of seconds to wait before deleting this response.
            **kwargs: Keyword arguments passed to ``CommandInteraction.create_initial_response`` or
                ``CommandInteraction.execute``.

        Returns:
            :obj:`~ResponseProxy`: Proxy wrapping the response of the ``respond`` call.

        .. versionadded:: 2.2.0
            ``delete_after`` kwarg.
        """

        async def _cleanup(timeout: t.Union[int, float], proxy_: ResponseProxy) -> None:
            await asyncio.sleep(timeout)

            try:
                await proxy_.delete()
            except hikari.NotFoundError:
                pass

        includes_ephemeral: t.Callable[
            [
                t.Union[hikari.MessageFlag, int],
            ],
            bool,
        ] = (
            lambda flags: (hikari.MessageFlag.EPHEMERAL & flags) == hikari.MessageFlag.EPHEMERAL
        )

        kwargs.pop("reply", None)
        kwargs.pop("mentions_reply", None)
        kwargs.pop("nonce", None)

        if (self._invoked or self._command).default_ephemeral:
            kwargs.setdefault("flags", hikari.MessageFlag.EPHEMERAL)

        if self._responded:
            kwargs.pop("response_type", None)
            if args and isinstance(args[0], hikari.ResponseType):
                args = args[1:]

            async def _ephemeral_followup_editor(
                _: ResponseProxy,
                *args_: t.Any,
                _wh_id: hikari.Snowflake,
                _tkn: str,
                _m_id: hikari.Snowflake,
                **kwargs_: t.Any,
            ) -> hikari.Message:
                return await self.app.rest.edit_webhook_message(_wh_id, _tkn, _m_id, *args_, **kwargs_)

            message = await self._interaction.execute(*args, **kwargs)
            proxy = ResponseProxy(
                message,
                editor=functools.partial(
                    _ephemeral_followup_editor,
                    _wh_id=self._interaction.webhook_id,
                    _tkn=self._interaction.token,
                    _m_id=message.id,
                ),
                deleteable=not includes_ephemeral(kwargs.get("flags", hikari.MessageFlag.NONE)),
            )
            self._responses.append(proxy)
            self._deferred = False

            if delete_after is not None:
                self.app.create_task(_cleanup(delete_after, proxy))

            return self._responses[-1]

        if args:
            if not isinstance(args[0], hikari.ResponseType):
                kwargs["content"] = args[0]
                kwargs.setdefault("response_type", hikari.ResponseType.MESSAGE_CREATE)
            else:
                kwargs["response_type"] = args[0]
                if len(args) > 1:
                    kwargs.setdefault("content", args[1])
        else:
            kwargs.setdefault("response_type", hikari.ResponseType.MESSAGE_CREATE)

        await self._interaction.create_initial_response(**kwargs)

        # Initial responses are special and need their own edit method defined
        # so that they work as expected for when the responses are ephemeral
        async def _editor(
            rp: ResponseProxy, *args_: t.Any, inter: hikari.CommandInteraction, **kwargs_: t.Any
        ) -> hikari.Message:
            await inter.edit_initial_response(*args_, **kwargs_)
            return await rp.message()

        proxy = ResponseProxy(
            fetcher=self._interaction.fetch_initial_response,
            editor=functools.partial(_editor, inter=self._interaction)
            if includes_ephemeral(kwargs.get("flags", hikari.MessageFlag.NONE))
            else None,
        )
        self._responses.append(proxy)
        self._responded = True

        if kwargs["response_type"] in (
            hikari.ResponseType.DEFERRED_MESSAGE_CREATE,
            hikari.ResponseType.DEFERRED_MESSAGE_UPDATE,
        ):
            self._deferred = True

        if delete_after is not None:
            self.app.create_task(_cleanup(delete_after, proxy))

        return self._responses[-1]

    async def respond_with_modal(
        self,
        title: str,
        custom_id: str,
        component: hikari.UndefinedOr[special_endpoints.ComponentBuilder] = hikari.UNDEFINED,
        components: hikari.UndefinedOr[t.Sequence[special_endpoints.ComponentBuilder]] = hikari.UNDEFINED,
    ) -> None:
        """
        Create a modal response to this context.

        Args:
            title (:obj:`str`): The title that will show up in the modal.
            custom_id (:obj:`str`): Developer set custom ID used for identifying interactions with this modal.
            component (UndefinedOr[:obj:`hikari.api.special_endpoints.ComponentBuilder`]):  A component builder
                to send in this modal.
            components (UndefinedOr[Sequence[:obj:`hikari.api.special_endpoints.ComponentBuilder`]]): A sequence
                of component builders to send in this modal.

        Returns:
            ``None``

        Raises:
            :obj:`ValueError`: If both ``component`` and ``components`` are specified or if neither are specified.

        .. versionadded:: 2.3.1
        """
        await self._interaction.create_modal_response(title, custom_id, component, components)
        self._responded = True



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["BaseHelpCommand", "DefaultHelpCommand", "filter_commands"]

import abc
import collections
import typing as t

from lightbulb import commands
from lightbulb import errors
from lightbulb.utils import nav

if t.TYPE_CHECKING:
    from lightbulb import app as app_
    from lightbulb import context as context_
    from lightbulb import plugins


async def filter_commands(
    cmds: t.Sequence[commands.base.Command], context: context_.base.Context
) -> t.Sequence[commands.base.Command]:
    """
    Evaluates the checks for each command provided, removing any that the checks fail for. This effectively
    removes any commands from the given collection that could not be invoked under the given context. This will
    also remove any commands with the ``hidden`` attribute set to ``True``.

    Args:
        cmds (Sequence[:obj:`~.commands.base.Command`]): Commands to filter.
        context (:obj:`~.context.base.Context`): Context to filter the commands under.

    Returns:
        Sequence[:obj:`~.commands.base.Command`]: Filtered commands.
    """
    new_cmds = []
    for cmd in cmds:
        if cmd.hidden:
            continue
        try:
            await cmd.evaluate_checks(context)
        except errors.CheckFailure:
            continue
        new_cmds.append(cmd)
    return new_cmds


class BaseHelpCommand(abc.ABC):
    """
    Base class for auto-generated help commands.

    Args:
        app (:obj:`~.app.BotApp`): The ``BotApp`` instance that the help command is registered to.
    """

    __slots__ = ("app",)

    def __init__(self, app: app_.BotApp) -> None:
        self.app: app_.BotApp = app
        """The ``BotApp`` instance the help command is registered to."""

    @property
    def bot(self) -> t.Optional[app_.BotApp]:
        """Alias for :obj:`~BaseHelpCommand.app`"""
        return self.app

    async def send_help(self, context: context_.base.Context, obj: t.Optional[str]) -> None:
        """
        Resolve the given object and send the help text for it to the given context.

        Help resolution order:

        - Prefix command
        - Slash command
        - Message command
        - User command
        - Plugin

        Args:
            context (:obj:`~.context.base.Context`): Context to send help to.
            obj (:obj:`str`): String representation of the object to send help for.
        """
        await self._send_help(context, obj)

    async def _send_help(self, context: context_.base.Context, obj: t.Optional[str]) -> None:
        if obj is None:
            await self.send_bot_help(context)
            return

        p_cmd = self.app.get_prefix_command(obj)
        if p_cmd is not None and not p_cmd.hidden:
            if isinstance(p_cmd, (commands.prefix.PrefixCommandGroup, commands.prefix.PrefixSubGroup)):
                await self.send_group_help(context, p_cmd)
                return
            await self.send_command_help(context, p_cmd)
            return
        s_cmd = self.app.get_slash_command(obj)
        if s_cmd is not None and not s_cmd.hidden:
            if isinstance(s_cmd, (commands.slash.SlashCommandGroup, commands.slash.SlashSubGroup)):
                await self.send_group_help(context, s_cmd)
                return
            await self.send_command_help(context, s_cmd)
            return
        m_cmd = self.app.get_message_command(obj)
        if m_cmd is not None and not m_cmd.hidden:
            await self.send_command_help(context, m_cmd)
            return
        u_cmd = self.app.get_user_command(obj)
        if u_cmd is not None and not u_cmd.hidden:
            await self.send_command_help(context, u_cmd)
            return

        plugin = self.app.get_plugin(obj)
        if plugin is not None:
            await self.send_plugin_help(context, plugin)
            return

        await self.object_not_found(context, obj)

    @abc.abstractmethod
    async def send_bot_help(self, context: context_.base.Context) -> None:
        """
        Sends an overall help message for the bot. This is called when no object is provided
        when the help command is invoked.

        Args:
            context (:obj:`~.context.base.Context`): Context to send help to.

        Returns:
            ``None``
        """
        ...

    @abc.abstractmethod
    async def send_command_help(self, context: context_.base.Context, command: commands.base.Command) -> None:
        """
        Sends a help message for the given command.

        Args:
            context (:obj:`~.context.base.Context`): Context to send help to.
            command (:obj:`~.commands.base.Command`): Command to send help for.

        Returns:
            ``None``
        """
        ...

    @abc.abstractmethod
    async def send_group_help(
        self,
        context: context_.base.Context,
        group: t.Union[
            commands.prefix.PrefixCommandGroup,
            commands.prefix.PrefixSubGroup,
            commands.slash.SlashCommandGroup,
            commands.slash.SlashSubGroup,
        ],
    ) -> None:
        """
        Sends a help message for the given command group.

        Args:
            context (:obj:`~.context.base.Context`): Context to send help to.
            group: Command group to send help for.

        Returns:
            ``None``
        """
        ...

    @abc.abstractmethod
    async def send_plugin_help(self, context: context_.base.Context, plugin: plugins.Plugin) -> None:
        """
        Sends a help message for the given plugin.

        Args:
            context (:obj:`~.context.base.Context`): Context to send help to.
            plugin (:obj:`~.plugins.Plugin`): Plugin to send help for.

        Returns:
            ``None``
        """
        ...

    async def object_not_found(self, context: context_.base.Context, obj: str) -> None:
        """
        Method called when no object could be resolved from the given name.

        Args:
            context (:obj:`~.context.base.Context`): Context to send help to.
            obj (:obj:`str`): String that the help command was invoked with but that could not be resolved
                into an object.

        Returns:
            ``None``
        """
        await context.respond(f"No command or category with the name `{obj}` could be found.")


class DefaultHelpCommand(BaseHelpCommand):
    """
    An implementation of the :obj:`~BaseHelpCommand` that the bot uses by default.
    """

    @staticmethod
    async def _get_command_plugin_map(
        cmd_map: t.Mapping[str, commands.base.Command], context: context_.base.Context
    ) -> t.Dict[t.Optional[plugins.Plugin], t.List[commands.base.Command]]:
        out = collections.defaultdict(list)
        for cmd in cmd_map.values():
            if await filter_commands([cmd], context):
                out[cmd.plugin].append(cmd)
        return out

    @staticmethod
    def _add_cmds_to_plugin_pages(
        pages: t.MutableMapping[t.Optional[plugins.Plugin], t.List[str]],
        cmds: t.Mapping[t.Optional[plugins.Plugin], t.List[commands.base.Command]],
        header: str,
    ) -> None:
        for plugin, cmds in cmds.items():
            pages[plugin].append(f"== {header} Commands")
            for cmd in set(cmds):
                pages[plugin].append(f"- {cmd.name} - {cmd.description}")

    async def send_bot_help(self, context: context_.base.Context) -> None:
        pages = []
        lines = [
            ">>> ```adoc",
            "==== Bot Help ====",
            "",
            f"For more information: {context.prefix}help [command|category]",
            "",
            "==== Categories ====",
        ]

        p_commands = await self._get_command_plugin_map(self.app._prefix_commands, context)
        s_commands = await self._get_command_plugin_map(self.app._slash_commands, context)
        m_commands = await self._get_command_plugin_map(self.app._message_commands, context)
        u_commands = await self._get_command_plugin_map(self.app._user_commands, context)

        plugin_pages: t.MutableMapping[t.Optional[plugins.Plugin], t.List[str]] = collections.defaultdict(list)
        self._add_cmds_to_plugin_pages(plugin_pages, p_commands, "Prefix")
        self._add_cmds_to_plugin_pages(plugin_pages, s_commands, "Slash")
        self._add_cmds_to_plugin_pages(plugin_pages, m_commands, "Message")
        self._add_cmds_to_plugin_pages(plugin_pages, u_commands, "User")

        for plugin, page in plugin_pages.items():
            if plugin is not None:
                lines.append(f"- {plugin.name}")
            pages.append(
                "\n".join(
                    [
                        ">>> ```adoc",
                        f"==== {plugin.name if plugin is not None else 'Uncategorised'} ====",
                        (f"{plugin.description}\n" if plugin.description else "No description provided\n")
                        if plugin is not None
                        else "",
                        *page,
                        "```",
                    ]
                )
            )
        lines.append("```")
        pages.insert(0, "\n".join(lines))

        navigator = nav.ButtonNavigator(pages)
        await navigator.run(context)

    async def send_command_help(self, context: context_.base.Context, command: commands.base.Command) -> None:
        long_help = command.get_help(context)
        prefix = (
            context.prefix
            if isinstance(command, commands.prefix.PrefixCommand)
            else "/"
            if isinstance(command, commands.slash.SlashCommand)
            else "\N{THREE BUTTON MOUSE}"
        )

        lines = [
            ">>> ```adoc",
            "==== Command Help ====",
            f"{command.name} - {command.description}",
            "",
            f"Usage: {prefix}{command.signature}",
            "",
            long_help if long_help else "No additional details provided.",
            "```",
        ]
        await context.respond("\n".join(lines))

    async def send_group_help(
        self,
        context: context_.base.Context,
        group: t.Union[
            commands.prefix.PrefixCommandGroup,
            commands.prefix.PrefixSubGroup,
            commands.slash.SlashCommandGroup,
            commands.slash.SlashSubGroup,
        ],
    ) -> None:
        long_help = group.get_help(context)
        prefix = (
            context.prefix
            if isinstance(group, commands.prefix.PrefixCommand)
            else "/"
            if isinstance(group, commands.slash.SlashCommand)
            else "\N{THREE BUTTON MOUSE}"
        )

        usages = list(
            filter(
                None,
                [
                    f"{prefix}{group.signature}" if isinstance(group, commands.prefix.PrefixCommand) else None,
                    f"{prefix}{group.qualname} [subcommand]",
                ],
            )
        )
        usages[0] = f"Usage: {usages[0]}"
        if len(usages) > 1:
            usages[1] = f"Or: {usages[1]}"

        lines = [
            ">>> ```adoc",
            "==== Group Help ====",
            f"{group.name} - {group.description}",
            "",
            "\n".join(usages),
            "",
            long_help if long_help else "No additional details provided.",
            "",
        ]
        if group._subcommands:
            subcommands = await filter_commands(group._subcommands.values(), context)  # type: ignore
            lines.append("== Subcommands")
            for cmd in set(subcommands):
                lines.append(f"- {cmd.name} - {cmd.description}")
        lines.append("```")
        await context.respond("\n".join(lines))

    async def send_plugin_help(self, context: context_.base.Context, plugin: plugins.Plugin) -> None:
        lines = [
            ">>> ```adoc",
            "==== Category Help ====",
            f"{plugin.name} - {plugin.description or 'No description provided'}",
            "",
        ]
        p_cmds, s_cmds, m_cmds, u_cmds = [], [], [], []
        all_commands = await filter_commands(plugin._all_commands, context)
        for cmd in all_commands:
            if isinstance(cmd, commands.prefix.PrefixCommand):
                p_cmds.append(cmd)
            elif isinstance(cmd, commands.slash.SlashCommand):
                s_cmds.append(cmd)
            elif isinstance(cmd, commands.message.MessageCommand):
                m_cmds.append(cmd)
            elif isinstance(cmd, commands.user.UserCommand):
                u_cmds.append(cmd)

        cmds: t.List[t.Tuple[t.Sequence[commands.base.Command], str]] = [
            (p_cmds, "Prefix"),
            (s_cmds, "Slash"),
            (m_cmds, "Message"),
            (u_cmds, "User"),
        ]
        for cmd_list, header in cmds:
            if cmd_list:
                lines.append(f"== {header} Commands")
                for cmd in set(cmd_list):
                    lines.append(f"- {cmd.name} - {cmd.description}")
        lines.append("```")
        await context.respond("\n".join(lines))



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = [
    "ButtonNavigator",
    "ComponentButton",
    "DataStore",
    "EmbedPaginator",
    "Paginator",
    "ReactionButton",
    "ReactionNavigator",
    "StringPaginator",
    "find",
    "first_page",
    "get",
    "last_page",
    "next_page",
    "permissions_for",
    "permissions_in",
    "prev_page",
    "stop",
    "build_invite_url",
    "data_store",
    "nav",
    "pag",
    "permissions",
    "search",
]

import typing as t
from urllib import parse

import hikari

from lightbulb.utils import data_store
from lightbulb.utils import nav
from lightbulb.utils import pag
from lightbulb.utils import permissions
from lightbulb.utils import search
from lightbulb.utils.data_store import *
from lightbulb.utils.nav import *
from lightbulb.utils.pag import *
from lightbulb.utils.permissions import *
from lightbulb.utils.search import *

if t.TYPE_CHECKING:
    from lightbulb import app as app_


def build_invite_url(
    app: app_.BotApp,
    permissions: hikari.Permissions = hikari.Permissions.NONE,
    allow_commands: bool = False,
) -> str:
    """
    Build an invite link for the given :obj:`~.app.BotApp` with the given permissions and scopes. This
    adds the ``bot`` scope to the URL by default.

    Args:
        app (:obj:`~.app.BotApp`): :obj:`~.app.BotApp` instance to create the invite link for.
        permissions (:obj:`hikari.Permissions`): Permissions to add to the invite link. Defaults to
            :obj:`hikari.Permissions.NONE`.
        allow_commands (:obj:`bool`): Whether to add the `applications.commands` scope to the invite link. Defaults
            to ``False``.

    Returns:
        :obj:`str`: Created invite link.

    .. versionadded:: 2.2.0
    """
    if app.application is None:
        raise ValueError("Cannot build invite url when 'app.application' is None")

    raw_scopes = {"bot"}
    if allow_commands:
        raw_scopes.add("applications.commands")

    scopes = " ".join(raw_scopes)
    return "https://discord.com/api/oauth2/authorize?" + parse.urlencode(
        {"client_id": app.application.id, "permissions": int(permissions), "scope": scopes}
    )



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
import importlib
import sys

import lightbulb

sys.stderr.write(f"hikari-lightbulb ({lightbulb.__version__})\n")
importlib.import_module("hikari.__main__")



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.

__all__ = [
    "ApplicationCommand",
    "Command",
    "CommandLike",
    "MessageCommand",
    "OptionLike",
    "OptionModifier",
    "PrefixCommand",
    "PrefixCommandGroup",
    "PrefixGroupMixin",
    "PrefixSubCommand",
    "PrefixSubGroup",
    "SlashCommand",
    "SlashCommandGroup",
    "SlashGroupMixin",
    "SlashSubCommand",
    "SlashSubGroup",
    "SubCommandTrait",
    "UserCommand",
]

from lightbulb.commands import base
from lightbulb.commands import message
from lightbulb.commands import prefix
from lightbulb.commands import slash
from lightbulb.commands import user
from lightbulb.commands.base import *
from lightbulb.commands.message import *
from lightbulb.commands.prefix import *
from lightbulb.commands.slash import *
from lightbulb.commands.user import *



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = [
    "LightbulbEvent",
    "LightbulbStartedEvent",
    "CommandErrorEvent",
    "PrefixCommandErrorEvent",
    "PrefixCommandInvocationEvent",
    "PrefixCommandCompletionEvent",
    "SlashCommandErrorEvent",
    "CommandInvocationEvent",
    "SlashCommandInvocationEvent",
    "SlashCommandCompletionEvent",
    "MessageCommandErrorEvent",
    "CommandCompletionEvent",
    "MessageCommandCompletionEvent",
    "MessageCommandInvocationEvent",
    "UserCommandErrorEvent",
    "UserCommandCompletionEvent",
    "UserCommandInvocationEvent",
]

import abc
import types
import typing as t

import attr
import hikari

if t.TYPE_CHECKING:
    from lightbulb import app as app_
    from lightbulb import commands
    from lightbulb import context as context_
    from lightbulb import errors


@attr.s(slots=True, weakref_slot=False)
class LightbulbEvent(hikari.Event, abc.ABC):
    """
    The base class for all lightbulb events. Every event dispatched by lightbulb
    will be an instance of a subclass of this.
    """

    app: app_.BotApp = attr.ib()
    """BotApp instance for this event."""

    @property
    def bot(self) -> app_.BotApp:
        """BotApp instance for this event. Alias for :obj:`~LightbulbEvent.app`."""
        return self.app


@attr.s(slots=True, weakref_slot=False)
class LightbulbStartedEvent(LightbulbEvent):
    """
    Event dispatched after the application commands have been managed.

    .. versionadded:: 2.1.0
    """


@attr.s(slots=True, weakref_slot=False)
class CommandErrorEvent(LightbulbEvent, abc.ABC):
    """
    The base class for all command error events. A subclass of this event will be dispatched whenever
    an error is encountered before or during the invocation of a command.
    """

    exception: errors.LightbulbError = attr.ib()
    """The exception that this event was triggered for."""
    context: context_.base.Context = attr.ib()
    """The context that this event was triggered for."""

    @property
    def exc_info(
        self,
    ) -> t.Tuple[t.Type[errors.LightbulbError], errors.LightbulbError, t.Optional[types.TracebackType]]:
        """The exception triplet compatible with context managers and :mod:`traceback` helpers."""
        return type(self.exception), self.exception, self.exception.__traceback__


@attr.s(slots=True, weakref_slot=False)
class CommandInvocationEvent(LightbulbEvent, abc.ABC):
    """
    The base class for all command invocation events. A subclass of this event will be dispatched before
    any command is invoked.
    """

    command: commands.base.Command = attr.ib()
    """The command that this event was triggered for."""
    context: context_.base.Context = attr.ib()
    """The context that this event was triggered for."""


@attr.s(slots=True, weakref_slot=False)
class CommandCompletionEvent(LightbulbEvent, abc.ABC):
    """
    The base class for all command completion events. A subclass of this event will be dispatched after
    command invocation completes. This will not be dispatched if any exceptions occur during invocation.
    """

    command: commands.base.Command = attr.ib()
    """The command that this event was triggered for."""
    context: context_.base.Context = attr.ib()
    """The context that this event was triggered for."""


@attr.s(slots=True, weakref_slot=False)
class PrefixCommandErrorEvent(CommandErrorEvent):
    """Event dispatched when an error is encountered before or during the invocation of a prefix command."""


@attr.s(slots=True, weakref_slot=False)
class SlashCommandErrorEvent(CommandErrorEvent):
    """Event dispatched when an error is encountered before or during the invocation of a slash command."""


@attr.s(slots=True, weakref_slot=False)
class MessageCommandErrorEvent(CommandErrorEvent):
    """Event dispatched when an error is encountered before or during the invocation of a message command."""


@attr.s(slots=True, weakref_slot=False)
class UserCommandErrorEvent(CommandErrorEvent):
    """Event dispatched when an error is encountered before or during the invocation of a user command."""


@attr.s(slots=True, weakref_slot=False)
class PrefixCommandInvocationEvent(CommandInvocationEvent):
    """Event dispatched before the invocation of a prefix command."""


@attr.s(slots=True, weakref_slot=False)
class SlashCommandInvocationEvent(CommandInvocationEvent):
    """Event dispatched before the invocation of a slash command."""


@attr.s(slots=True, weakref_slot=False)
class MessageCommandInvocationEvent(CommandInvocationEvent):
    """Event dispatched before the invocation of a message command."""


@attr.s(slots=True, weakref_slot=False)
class UserCommandInvocationEvent(CommandInvocationEvent):
    """Event dispatched before the invocation of a user command."""


@attr.s(slots=True, weakref_slot=False)
class PrefixCommandCompletionEvent(CommandCompletionEvent):
    """Event dispatched after the invocation of a prefix command is completed."""


@attr.s(slots=True, weakref_slot=False)
class SlashCommandCompletionEvent(CommandCompletionEvent):
    """Event dispatched after the invocation of a slash command is completed."""


@attr.s(slots=True, weakref_slot=False)
class MessageCommandCompletionEvent(CommandCompletionEvent):
    """Event dispatched after the invocation of a message command is completed."""


@attr.s(slots=True, weakref_slot=False)
class UserCommandCompletionEvent(CommandCompletionEvent):
    """Event dispatched after the invocation of a user command is completed."""



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ["Trigger", "UniformTrigger", "CronTrigger"]

import abc
import datetime
import typing as t

try:
    import croniter

    _CRON_AVAILABLE = True

except ModuleNotFoundError:
    _CRON_AVAILABLE = False

_CT = t.Union[int, str, None]


class Trigger(abc.ABC):
    """
    Base class representing a task trigger.
    """

    __slots__ = ()

    @abc.abstractmethod
    def get_interval(self) -> float:
        """
        Get the interval to sleep for in seconds until the next run of the task.

        Returns:
            :obj:`float`: Seconds to sleep until the next time the task is run.
        """
        ...

    @property
    def wait_before_execution(self) -> bool:
        """
        Whether the trigger should wait `get_interval` amount of time before executing
        for the first time by default. Will be overridden by the `wait_before_execution`
        parameter in the `task` decorator.
        """
        return False


class UniformTrigger(Trigger):
    """
    A uniform interval task trigger. Tasks will always be run with the same interval between them.

    Args:
        interval (:obj:`float`): Number of seconds between task executions.
    """

    __slots__ = ("_interval",)

    def __init__(self, interval: float) -> None:
        self._interval = interval

    def get_interval(self) -> float:
        return self._interval


if t.TYPE_CHECKING or _CRON_AVAILABLE:

    class CronTrigger(Trigger):
        """
        A crontab-based task trigger. Tasks will be run dependent on the given crontab.

        Keyword arguments are only used if no crontab expression is passed.

        Args:
            crontab (Optional[:obj:`str`]): Schedule that task executions will follow.

        Keyword Args:
            month (Optional[Union[:obj:`int`, :obj:`str`]]):
                The month(s) that the task will be executed.
            day (Optional[Union[:obj:`int`, :obj:`str`]]):
                The day(s) that the task will be executed.
            day_of_week (Optional[Union[:obj:`int`, :obj:`str`]]):
                The day(s) of the week that the task will be executed.
            hour (Optional[Union[:obj:`int`, :obj:`str`]]):
                The hour(s) that the task will be executed.
            minute (Optional[Union[:obj:`int`, :obj:`str`]]):
                The minute(s) that the task will be executed.
            second (Optional[Union[:obj:`int`, :obj:`str`]]):
                The second(s) that the task will be executed.

        Note:
            This trigger is only available when installing lightbulb with the ``[crontrigger]`` option.

            See: ``pip install hikari-lightbulb[crontrigger]``
        """

        __slots__ = ("_croniter",)

        @t.overload
        def __init__(self, crontab: str, /) -> None:
            ...

        @t.overload
        def __init__(
            self, /, *, minute: _CT = "*", hour: _CT = "*", month: _CT = "*", day_of_week: _CT = "*", second: _CT = 0
        ) -> None:
            ...

        def __init__(self, crontab: t.Optional[str] = None, /, **kwargs: _CT) -> None:
            if not crontab:
                crontab = f"{kwargs.get('minute', '*')} {kwargs.get('hour', '*')} {kwargs.get('day', '*')} {kwargs.get('month', '*')} {kwargs.get('day_of_week', '*')} {kwargs.get('second', 0)}"

            self._croniter = croniter.croniter(crontab, datetime.datetime.now(datetime.timezone.utc))

        def get_interval(self) -> float:
            difference: datetime.timedelta = self._croniter.get_next(datetime.datetime) - datetime.datetime.now(
                datetime.timezone.utc
            )
            return difference.total_seconds()

        @property
        def wait_before_execution(self) -> bool:
            return True

else:

    class CronTrigger(Trigger):
        __slots__ = ()

        def __init__(self, _: str, /, **__: _CT) -> None:
            raise ModuleNotFoundError(
                "'CronTrigger' is not available. Install lightbulb with option '[crontrigger]' to enable"
            )

        def get_interval(self) -> float:
            raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = [
    "implements",
    "command",
    "option",
    "add_checks",
    "check_exempt",
    "add_cooldown",
    "set_help",
    "set_max_concurrency",
    "app_command_permissions",
]

import functools
import inspect
import typing as t

import hikari

from lightbulb import buckets
from lightbulb import checks as checks_
from lightbulb import commands
from lightbulb import cooldown_algorithms
from lightbulb import cooldowns

if t.TYPE_CHECKING:
    from lightbulb import context

T = t.TypeVar("T")
CommandCallbackT = t.TypeVar("CommandCallbackT", bound=t.Callable[..., t.Coroutine[t.Any, t.Any, None]])
AutocompleteCallbackT = t.Callable[
    [hikari.CommandInteractionOption, hikari.AutocompleteInteraction],
    t.Coroutine[
        t.Any,
        t.Any,
        t.Union[
            str, hikari.api.AutocompleteChoiceBuilder, t.Sequence[t.Union[str, hikari.api.AutocompleteChoiceBuilder]]
        ],
    ],
]


def implements(
    *command_types: t.Type[commands.base.Command],
) -> t.Callable[[CommandCallbackT], CommandCallbackT]:
    """
    Second order decorator that defines the command types that a given callback function will implement.

    Args:
        *command_types (Type[:obj:`~.commands.base.Command`]): Command types that the function will implement.
    """

    def decorate(func: CommandCallbackT) -> CommandCallbackT:
        setattr(func, "__cmd_types__", command_types)
        return func

    return decorate


def command(
    name: str, description: str, *, cls: t.Type[commands.base.CommandLike] = commands.base.CommandLike, **kwargs: t.Any
) -> t.Callable[[CommandCallbackT], commands.base.CommandLike]:
    """
    Second order decorator that converts the decorated function into a :obj:`~.commands.base.CommandLike` object.

    Args:
        name (:obj:`str`): The name of the command .
        description (:obj:`str`): The description of the command.

    Keyword Args:
        error_handler (Optional[ListenerT]): The function to register as the command's error handler. Defaults to
            ``None``. This can also be set with the :obj:`~.commands.base.CommandLike.set_error_handler`
            decorator.
        aliases (Sequence[:obj:`str`]): Aliases for the command. This will only affect prefix commands. Defaults
            to an empty list.
        guilds (Sequence[:obj:`int`]): The guilds that the command will be created in. This will only affect
            application commands. Defaults to an empty list.
        parser (:obj:`~.utils.parser.BaseParser`): The argument parser to use for prefix commands. Defaults
            to :obj:`~.utils.parser.Parser`.
        auto_defer (:obj:`bool`): Whether to automatically defer the response when the command is invoked. If
            ``True``, the bot will send an initial response of type ``DEFERRED_MESSAGE_CREATE`` for interactions, and
            for prefix commands, typing will be triggered in the invocation channel.
        ephemeral (:obj:`bool`): Whether to send responses from the invocation of this command as ephemeral by
            default. If ``True`` then all responses from the command will use the flag :obj:`hikari.MessageFlags.EPHEMERAL`.
            This will not affect prefix commands as responses from prefix commands **cannot** be ephemeral.
        hidden (:obj:`bool`): Whether to hide the command from the help command. Defaults to ``False``.
        inherit_checks (:obj:`bool`): Whether or not the command should inherit checks from the parent group. Only
            affects subcommands. Defaults to ``False``.
        pass_options (:obj:`bool`): Whether the command will have its options passed as keyword arguments when invoked.
        name_localizations (Mapping[Union[:obj:`hikari.locales.Locale`, :obj:`str`], :obj:`str`]): A mapping of
            locale to name localizations for this command. Defaults to an empty dictionary.
        description_localizations (Mapping[Union[:obj:`hikari.locales.Locale`, :obj:`str`], :obj:`str`]): A mapping
            of locale to description localizations for this command. Defaults to an empty dictionary.
        nsfw (:obj:`bool`): Whether the command should only be enabled in NSFW channels. For prefix commands, this
            will add an NSFW-channel only check to the command automatically. For slash commands, this will behave
            as specified in the Discord documentation.
        cls (Type[:obj:`~.commands.base.CommandLike`]): ``CommandLike`` class to instantiate from this decorator.
            Defaults to :obj:`~.commands.base.CommandLike`.

    .. versionadded:: 2.1.0
        ``cls`` kwarg.
    .. versionadded:: 2.2.1
        ``pass_options`` kwarg.
    .. versionadded:: 2.3.0
        ``name_localizations`` and ``description_localizations`` kwargs.
    """

    def decorate(func: CommandCallbackT) -> commands.base.CommandLike:
        cmd = cls(func, name, description, **kwargs)
        if cmd.nsfw:
            # Add nsfw channel check only for prefix commands. For other command types the check will just return True
            add_checks(checks_.Check(checks_._nsfw_channel_only, lambda ctx: True, lambda ctx: True, lambda ctx: True))(
                cmd
            )
        return cmd

    return decorate


def option(
    name: str,
    description: str,
    type: t.Any = str,
    *,
    required: hikari.UndefinedOr[bool] = hikari.UNDEFINED,
    choices: t.Optional[t.Sequence[t.Union[str, int, float, hikari.CommandChoice]]] = None,
    channel_types: t.Optional[t.Sequence[hikari.ChannelType]] = None,
    default: hikari.UndefinedOr[t.Any] = hikari.UNDEFINED,
    modifier: commands.base.OptionModifier = commands.base.OptionModifier.NONE,
    min_value: t.Optional[t.Union[int, float]] = None,
    max_value: t.Optional[t.Union[int, float]] = None,
    min_length: t.Optional[int] = None,
    max_length: t.Optional[int] = None,
    autocomplete: t.Union[bool, AutocompleteCallbackT] = False,
    cls: t.Type[commands.base.OptionLike] = commands.base.OptionLike,
    **kwargs: t.Any,
) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
    """
    Second order decorator that adds an option to the decorated :obj:`~.commands.base.CommandLike`
    object.

    Args:
        name (:obj:`str`): The name of the option.
        description (:obj:`str`): The description of the option.
        type (Any): The type of the option. This will be used as the converter for prefix commands.

    Keyword Args:
        required (UndefinedOr[:obj:`bool`]): Whether this option is required. This will be inferred from whether a
            default value for the option was specified if not otherwise specified.
        choices (Optional[Sequence[Union[:obj:`str`, :obj:`int`, :obj:`float`, :obj:`~hikari.commands.CommandChoice`]]]): The
            choices for the option. This will only affect slash commands. Defaults to ``None``.
        channel_types (Optional[Sequence[hikari.channels.ChannelType]]): The channel types allowed for the option.
            This will only affect slash commands. Defaults to ``None``.
        default (UndefinedOr[Any]): The default value for the option. Defaults to :obj:`~hikari.undefined.UNDEFINED`.
        modifier (:obj:`~.commands.base.OptionModifier`): Modifier controlling how the option should be parsed. Defaults
            to ``OptionModifier.NONE``.
        min_value (Optional[Union[:obj:`float`, :obj:`int`]]): The minimum value permitted for this option (inclusive).
            Only available if the option type is numeric (integer or float). Defaults to ``None``.
        max_value (Optional[Union[:obj:`float`, :obj:`int`]]): The maximum value permitted for this option (inclusive).
            Only available if the option type is numeric (integer or float). Defaults to ``None``.
        min_length (Optional[:obj:`int`]): The minimum length permitted for this option. Only available if this option
            is string type.
        max_length (Optional[:obj:`int`]): The maximum length permitted for this option. Only available if this option
            is string type.
        autocomplete (Union[:obj:`bool`, AutocompleteCallbackT]): Boolean representing whether the option will use
            autocomplete or the callback to use for autocomplete for this option. This will only affect slash commands.
            Defaults to ``False``.
        name_localizations (Mapping[Union[:obj:`hikari.locales.Locale`, :obj:`str`], :obj:`str`]): A mapping of locale
            to name localizations for this option. Defaults to an empty dictionary.
        description_localizations (Mapping[Union[:obj:`hikari.locales.Locale`, :obj:`str`], :obj:`str`]): A mapping of
            locale to description localizations for this option. Defaults to an empty dictionary.
        cls (Type[:obj:`~.commands.base.OptionLike`]): ``OptionLike`` class to instantiate from this decorator. Defaults
            to :obj:`~.commands.base.OptionLike`.

    .. versionadded:: 2.1.0
        ``cls`` kwarg.
    .. versionadded:: 2.1.3
        ``min_value`` and ``max_value`` kwargs.
    .. versionadded:: 2.3.2
        ``min_length`` and ``max_length`` kwargs.
    """

    def decorate(c_like: commands.base.CommandLike) -> commands.base.CommandLike:
        nonlocal default, required, autocomplete

        if not isinstance(c_like, commands.base.CommandLike):
            raise SyntaxError("'option' decorator must be above the 'command' decorator")

        required = required if required is not hikari.UNDEFINED else (default is hikari.UNDEFINED)
        assert isinstance(required, bool)
        if not required:
            default = None if default is hikari.UNDEFINED else default

        if not isinstance(autocomplete, bool):
            c_like._autocomplete_callbacks[name] = autocomplete
            autocomplete = True

        c_like.options[name] = cls(
            name=name,
            description=description,
            arg_type=type,
            required=required,
            choices=choices,
            channel_types=channel_types,
            default=default,
            modifier=modifier,
            min_value=min_value,
            max_value=max_value,
            min_length=min_length,
            max_length=max_length,
            autocomplete=autocomplete,
            **kwargs,
        )
        return c_like

    return decorate


def add_checks(
    *cmd_checks: t.Union[checks_.Check, checks_._ExclusiveCheck]
) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
    """
    Second order decorator that adds one or more checks to the decorated :obj:`~.commands.base.CommandLike`
    object.

    Args:
        *cmd_checks (:obj:`~.checks.Check`): Check object(s) to add to the command.
    """

    def decorate(c_like: commands.base.CommandLike) -> commands.base.CommandLike:
        if not isinstance(c_like, commands.base.CommandLike):
            raise SyntaxError("'add_checks' decorator must be above the 'command' decorator")

        new_checks = [*c_like.checks, *cmd_checks]
        c_like.checks = new_checks

        for check in cmd_checks:
            check.add_to_object_hook(c_like)

        return c_like

    return decorate


def check_exempt(
    predicate: t.Callable[[context.base.Context], t.Union[bool, t.Coroutine[t.Any, t.Any, bool]]]
) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
    """
    Second order decorator which allows all checks to be bypassed if the ``predicate`` conditions are met.
    Predicate can be a synchronous or asynchronous function but must take a single argument - ``context`` - and
    must return  a boolean - ``True`` if checks should be bypassed or ``False`` if not.

    Args:
        predicate (Callable[[:obj:`~.context.base.Context`], Union[:obj:`bool`, Coroutine[Any, Any, :obj:`bool`]]): The
            function to call to check if the checks should be bypassed for the given context.
    """

    def decorate(c_like: commands.base.CommandLike) -> commands.base.CommandLike:
        if not isinstance(c_like, commands.base.CommandLike):
            raise SyntaxError("'check_exempt' decorator must be above the 'command' decorator")

        c_like.check_exempt = predicate
        return c_like

    return decorate


@t.overload
def add_cooldown(
    length: float,
    uses: int,
    bucket: t.Type[buckets.Bucket],
    algorithm: t.Type[cooldown_algorithms.CooldownAlgorithm] = cooldown_algorithms.BangBangCooldownAlgorithm,
) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
    ...


@t.overload
def add_cooldown(
    *,
    callback: t.Callable[[context.base.Context], t.Union[buckets.Bucket, t.Coroutine[t.Any, t.Any, buckets.Bucket]]],
) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
    ...


def add_cooldown(
    length: t.Optional[float] = None,
    uses: t.Optional[int] = None,
    bucket: t.Optional[t.Type[buckets.Bucket]] = None,
    algorithm: t.Type[cooldown_algorithms.CooldownAlgorithm] = cooldown_algorithms.BangBangCooldownAlgorithm,
    *,
    callback: t.Optional[
        t.Callable[[context.base.Context], t.Union[buckets.Bucket, t.Coroutine[t.Any, t.Any, buckets.Bucket]]]
    ] = None,
    cls: t.Type[cooldowns.CooldownManager] = cooldowns.CooldownManager,
) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
    """
    Second order decorator that sets the cooldown manager for a command.

    Args:
        length (:obj:`float`): The length in seconds of the cooldown timer.
        uses (:obj:`int`): The number of command invocations before the cooldown will be triggered.
        bucket (Type[:obj:`~.buckets.Bucket`]): The bucket to use for cooldowns.
        algorithm (Type[:obj:`~.cooldown_algorithms.CooldownAlgorithm`]): The cooldown algorithm to use. Defaults
            to :obj:`~.cooldown_algorithms.BangBangCooldownAlgorithm`.

    Keyword Args:
        callback (Callable[[:obj:`~.context.base.Context], Union[:obj:`~.buckets.Bucket`, Coroutine[Any, Any, :obj:`~.buckets.Bucket`]]]): Callable
            that takes the context the command was invoked under and returns the appropriate bucket object to use for
            cooldowns in the context.
        cls (Type[:obj:`~.cooldowns.CooldownManager`]): The cooldown manager class to use. Defaults to
            :obj:`~.cooldowns.CooldownManager`.
    """
    getter: t.Callable[[context.base.Context], t.Union[buckets.Bucket, t.Coroutine[t.Any, t.Any, buckets.Bucket]]]
    if length is not None and uses is not None and bucket is not None:

        def _get_bucket(
            _: context.base.Context,
            b: t.Type[buckets.Bucket],
            l: float,
            u: int,
            a: t.Type[cooldown_algorithms.CooldownAlgorithm],
        ) -> buckets.Bucket:
            return b(l, u, a)

        getter = functools.partial(_get_bucket, b=bucket, l=length, u=uses, a=algorithm)
    elif callback is not None:
        getter = callback
    else:
        raise TypeError("Invalid arguments - either provided all of the args length,uses,bucket or the kwarg callback")

    def decorate(c_like: commands.base.CommandLike) -> commands.base.CommandLike:
        if not isinstance(c_like, commands.base.CommandLike):
            raise SyntaxError("'add_cooldown' decorator must be above the 'command' decorator")

        c_like.cooldown_manager = cls(getter)
        return c_like

    return decorate


def set_help(
    text: t.Optional[t.Union[str, t.Callable[[commands.base.Command, context.base.Context], str]]] = None,
    *,
    docstring: bool = False,
) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
    """
    Second order decorator that defines the long help text for a command, or how the long help
    text should be retrieved. If ``text`` is provided then it will override the value for ``docstring``.

    Args:
        text (Union[Callable[[:obj:`~.commands.base.Command`, :obj:`~.context.base.Context`], :obj:`str`], :obj:`str`]): The
            long help text for the command, or a **synchronous** function called with the :obj:`~.commands.base.Command`
            object to get help text for and the :obj:`~.context.base.Context` that the help text should be
            retrieved for. If this is not provided, then you **must** pass the kwarg ``docstring=True``.

    Keyword Args:
        docstring (:obj:`bool`): Whether the command help text should be extracted from the command's docstring.
            If this is ``False`` (default) then a value **must** be provided for the ``text`` arg.
    """
    if text is None and docstring is False:
        raise ValueError("Either help text/callable or docstring=True must be provided")

    def decorate(c_like: commands.base.CommandLike) -> commands.base.CommandLike:
        if not isinstance(c_like, commands.base.CommandLike):
            raise SyntaxError("'set_help' decorator must be above the 'command' decorator")

        if isinstance(text, str):
            getter = lambda _, __: text
        elif docstring:
            cmd_doc = inspect.getdoc(c_like.callback)
            if cmd_doc is None:
                raise ValueError("docstring=True was provided but the command does not have a docstring")
            getter = lambda _, __: cmd_doc  # type: ignore
        else:
            assert text is not None
            getter = text

        c_like.help_getter = getter
        return c_like

    return decorate


def set_max_concurrency(
    uses: int, bucket: t.Type[buckets.Bucket]
) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
    """
    Second order decorator that defines the max concurrency limit for a command.

    Args:
        uses (:obj:`int`): The maximum number of uses of the command that can be executing
            concurrently before a :obj:`~.errors.MaxConcurrencyLimitReached` will be raised upon invocation.
        bucket (Type[:obj:`~.buckets.Bucket`]): Bucket that command max concurrency will be processed under.

    .. versionadded:: 2.2.1
    """
    if uses < 1 or not isinstance(uses, int):
        raise ValueError("'uses' must be a positive integer")

    def decorate(c_like: commands.base.CommandLike) -> commands.base.CommandLike:
        if not isinstance(c_like, commands.base.CommandLike):
            raise SyntaxError("'set_max_concurrency' decorator must be above the 'command' decorator")
        c_like.max_concurrency = (uses, bucket)

        return c_like

    return decorate


def app_command_permissions(
    perms: t.Optional[hikari.Permissions] = None, *, dm_enabled: bool = True, bypass_checks: bool = False
) -> t.Callable[[commands.base.CommandLike], commands.base.CommandLike]:
    """
    Second order decorator that defines the application command permissions v2 configuration
    for a command.

    Args:
        perms (Optional[:obj:`hikari.Permissions`]): The permissions to set as the default for the command when
            it is created in a guild. If ``None``, no default permissions will be set.
        dm_enabled (:obj:`bool`): Whether the command should be available in DM. If the command is not global
            then this will have no effect.
        bypass_checks (:obj:`bool`): Whether author permissions checks will be bypassed when this command is
            invoked as an application command. This allows guild admins to dictate who is allowed to use the command
            as opposed to enforcing a set of permissions.

    .. versionadded:: 2.2.3
    """

    def decorate(c_like: commands.base.CommandLike) -> commands.base.CommandLike:
        if perms is not None:
            c_like.app_command_default_member_permissions = perms

        c_like.app_command_dm_enabled = dm_enabled
        c_like.app_command_bypass_author_permission_checks = bypass_checks

        return c_like

    return decorate



# -*- coding: utf-8 -*-
# Copyright © tandemdude 2020-present
#
# This file is part of Lightbulb.
#
# Lightbulb is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Lightbulb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Lightbulb. If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = [
    "LightbulbError",
    "UnsupportedResponseOperation",
    "ApplicationCommandCreationFailed",
    "CommandNotFound",
    "CommandInvocationError",
    "CommandIsOnCooldown",
    "ConverterFailure",
    "NotEnoughArguments",
    "MissingRequiredAttachmentArgument",
    "MaxConcurrencyLimitReached",
    "CheckFailure",
    "InsufficientCache",
    "NotOwner",
    "OnlyInGuild",
    "OnlyInDM",
    "BotOnly",
    "WebhookOnly",
    "HumanOnly",
    "NSFWChannelOnly",
    "ExtensionMissingUnload",
    "ExtensionNotFound",
    "ExtensionNotLoaded",
    "ExtensionMissingLoad",
    "ExtensionAlreadyLoaded",
    "CommandAlreadyExists",
    "MissingRequiredRole",
    "MissingRequiredPermission",
    "BotMissingRequiredPermission",
    "MissingRequiredAttachment",
]

import typing as t
import warnings

import hikari

if t.TYPE_CHECKING:
    from lightbulb import buckets
    from lightbulb import commands


class LightbulbError(Exception):
    """
    Base lightbulb exception class. All errors raised by lightbulb will be a subclass
    of this exception.
    """


class UnsupportedResponseOperation(LightbulbError):
    """
    Exception raised when ``edit`` or ``delete`` is called on a response that cannot be
    edited and/or deleted (generally occurs for ephemeral responses).
    """


class ApplicationCommandCreationFailed(LightbulbError):
    """Exception raised when initialisation of application commands fails."""


class ExtensionNotFound(LightbulbError):
    """Exception raised when an attempt is made to load an extension that does not exist."""


class ExtensionAlreadyLoaded(LightbulbError):
    """Exception raised when an extension already loaded is attempted to be loaded."""


class ExtensionMissingLoad(LightbulbError):
    """Exception raised when an extension is attempted to be loaded but does not contain a load function"""


class ExtensionMissingUnload(LightbulbError):
    """Exception raised when an extension is attempted to be unloaded but does not contain an unload function"""


class ExtensionNotLoaded(LightbulbError):
    """Exception raised when an extension not already loaded is attempted to be unloaded."""


class CommandAlreadyExists(LightbulbError):
    """
    Error raised when attempting to add a command to the bot but a name or alias
    for the command conflicts with a command that already exists.
    """


class CommandNotFound(LightbulbError):
    """
    Error raised when a command is attempted to be invoked but an implementation
    is not found. This will only be raised for prefix commands.
    """

    __slots__ = ("invoked_with",)

    def __init__(self, *args: t.Any, invoked_with: str) -> None:
        super().__init__(*args)
        self.invoked_with: str = invoked_with
        """The name or alias of the command that was used."""


class CommandInvocationError(LightbulbError):
    """
    Error raised when an error is encountered during command invocation. This
    wraps the original exception that caused it, which is accessible through
    ``CommandInvocationError.__cause__`` or ``CommandInvocationError.original``.
    """

    __slots__ = ("original",)

    def __init__(self, *args: t.Any, original: Exception) -> None:
        super().__init__(*args)
        self.original: Exception = original
        """The exception that caused this to be raised. Also accessible through ``CommandInvocationError.__cause__``"""
        self.__cause__ = original


class CommandIsOnCooldown(LightbulbError):
    """
    Error raised when a command was on cooldown when it was attempted to be invoked.
    """

    __slots__ = ("retry_after",)

    def __init__(self, *args: t.Any, retry_after: float) -> None:
        super().__init__(*args)
        self.retry_after: float = retry_after
        """The amount of time in seconds remaining until the cooldown expires."""


class ConverterFailure(LightbulbError):
    """
    Error raised when option type conversion fails while prefix command arguments are being parsed.
    """

    __slots__ = ("option", "raw_value")

    def __init__(self, *args: t.Any, opt: commands.base.OptionLike, raw: str) -> None:
        super().__init__(*args)
        self.option: commands.base.OptionLike = opt
        """The option that could not be converted."""
        self.raw_value: str = raw
        """
        The value that could not be converted.

        .. versionadded:: 2.2.1
        """


class NotEnoughArguments(LightbulbError):
    """
    Error raised when a prefix command expects more options than could be parsed from the user's input.
    """

    __slots__ = ("missing_options",)

    def __init__(self, *args: t.Any, missing: t.Sequence[commands.base.OptionLike]) -> None:
        super().__init__(*args)
        self.missing_options: t.Sequence[commands.base.OptionLike] = missing
        """The missing options from the command invocation."""


class MissingRequiredAttachmentArgument(LightbulbError):
    """
    Error raised when a prefix command expects an attachment but none were supplied with the invocation.
    """

    __slots__ = ("missing_option",)

    def __init__(self, *args: t.Any, missing: commands.base.OptionLike) -> None:
        super().__init__(*args)
        self.missing_option: commands.base.OptionLike = missing
        """The missing attachment option from the command invocation."""


class MaxConcurrencyLimitReached(LightbulbError):
    """
    Error raised when the maximum number of allowed concurrent invocations for a command
    has been exceeded.
    """

    __slots__ = ("bucket",)

    def __init__(self, *args: t.Any, bucket: t.Type[buckets.Bucket]) -> None:
        super().__init__(*args)
        self.bucket = bucket
        """
        The bucket type that triggered the max concurrency limit.

        .. versionadded:: 2.3.3
        """


class CheckFailure(LightbulbError):
    """
    Error raised when a check fails before command invocation. If another error caused this
    to be raised then you can access it using ``CheckFailure.__cause__``, or in the case of
    multiple checks failing, via ``CheckFailure.causes`` (since version `2.2.1`).
    """

    __slots__ = ("causes",)

    def __init__(self, *args: t.Any, causes: t.Optional[t.Sequence[Exception]] = None) -> None:
        super().__init__(*args)
        self.causes: t.Sequence[Exception] = causes or []


class InsufficientCache(CheckFailure):
    """
    Error raised when the cache is required for an operation but either could not be accessed
    or did not return the required object.
    """


class NotOwner(CheckFailure):
    """
    Error raised when a user who is not the owner of the bot attempts to use a command
    that is restricted to owners only.
    """


class OnlyInGuild(CheckFailure):
    """
    Error raised when a user attempts to use a command in DMs that has been restricted
    to being used only in guilds.
    """


class OnlyInDM(CheckFailure):
    """
    Error raised when a user attempts to use a command in a guild that has been restricted
    to being used only in DMs.
    """


class BotOnly(CheckFailure):
    """
    Error raised when any entity other than a bot attempts to use a command that has been
    restricted to being used only by bots.
    """


class WebhookOnly(CheckFailure):
    """
    Error raised when any entity other than a webhook attempts to use a command that has been
    restricted to being used only by webhooks.
    """


class HumanOnly(CheckFailure):
    """
    Error raised when any entity other than a human attempts to use a command that has been
    restricted to being used only by humans.
    """


class NSFWChannelOnly(CheckFailure):
    """
    Error raised when a user attempts to use a command in a non-NSFW channel that has
    been restricted to only being used in NSFW channels.
    """


class MissingRequiredRole(CheckFailure):
    """
    Error raised when the member invoking a command is missing one or more of the required roles.
    """

    def __init__(self, *args: t.Any, roles: t.Sequence[int], mode: t.Callable[[t.Sequence[bool]], bool]) -> None:
        super().__init__(*args)
        self.missing_roles: t.Sequence[int] = roles
        "The roles that the member is missing."
        self.mode = mode
        "The mode used to check the roles, can be either ``any`` or ``all``."


class MissingRequiredPermission(CheckFailure):
    """
    Error raised when the member invoking a command is missing one or more of the required permissions
    in order to be able to run the command.
    """

    def __init__(self, *args: t.Any, perms: hikari.Permissions) -> None:
        super().__init__(*args)
        self.missing_perms: hikari.Permissions = perms
        """The permissions that the member is missing."""


class BotMissingRequiredPermission(CheckFailure):
    """
    Error raised when the bot is missing one or more of the required permissions
    in order to be able to run the command.
    """

    def __init__(self, *args: t.Any, perms: hikari.Permissions) -> None:
        super().__init__(*args)
        self.missing_perms: hikari.Permissions = perms
        """The permissions that the bot is missing."""


class MissingRequiredAttachment(CheckFailure):
    """
    Error raised when an attachment is required for the command but none were supplied with the invocation.
    """

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        warnings.warn(
            "'MissingRequiredAttachment' is deprecated and scheduled for removal in version '2.5.0'. "
            "Use an option with type 'hikari.Attachment' and catch 'MissingRequiredAttachmentArgument' instead.",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)
